import os
import rasterio

class Raster(object):
    """
    A class to represent a raster object.
    
    """
    
    def __init__(self, source, iteration=None, timestep=None):
        self.__source = source
        self.__name = self.__init_name(iteration, timestep)
        self.__dimensions = self.__init_dimensions()
        self.__extent = self.__init_extent()
        self.__resolution = self.__init_resolution()
        self.__crs = self.__init_crs()
        
    def __str__(self):
        
        return self.__to_string()
    
    def __repr__(self):
        
        return self.__to_string()
    
    def __to_string(self):
        
        s = "class: Raster\n"
        s += "source: %s\n" % self.source
        s += "name: %s\n" % self.name
        s += "dimensions: %s\n" % self.dimensions
        s += "resolution: %s\n" % self.resolution
        s += "extent: %s\n" % self.extent
        s += "crs: %s\n" % self.crs
        
        return s
        
    @property
    def source(self):
        """Gets the filepath of the raster"""
        return self.__source
    
    @property
    def name(self):
        """Gets the name of the raster"""
        return self.__name
    
    @property
    def dimensions(self):
        """Gets the dimensions of the raster"""
        return self.__dimensions
    
    @property
    def resolution(self):
        """Gets the resolution of the raster"""
        return self.__resolution
    
    @property
    def extent(self):
        """Gets the extent of the raster"""
        return self.__extent
    
    @property
    def crs(self):
        """Gets the coordinate system of the raster"""
        return self.__crs
    
    def values(self, band=None):
        """
        Gets the values in each cell of the raster
        

        Parameters
        ----------
        band : Int, optional
            The specific band to return. If None, then all bands in the raster
            are returned in the array, with the first dimension of the array
            corresponding to the band number. The default is None.

        Returns
        -------
        values : numpy array
            Array of values corresponding to the cell values in the raster.

        """
        
        if band is None:
            with rasterio.open(self.source) as raster:
                values = raster.read()
                if values.shape[0] == 1:
                    values = values[0]
            return values
        
        else:
            with rasterio.open(self.source) as raster:
                values = raster.read(band)
            return values
        
    def __init_name(self, iteration, timestep):
        
        prefix = os.path.splitext(os.path.basename(self.source))[0]
        
        if prefix.endswith(f".it{iteration}.ts{timestep}"):
            return prefix
        else:
            return prefix + ".it" + str(iteration) + ".ts" + str(timestep)
    
    def __init_dimensions(self):
        
        dim_dict = {"height": [], "width": [], "cells": []}
        
        with rasterio.open(self.source) as raster:
            
            dim_dict["height"].append(raster.height)
            dim_dict["width"].append(raster.width)
            dim_dict["cells"].append(raster.height * raster.width)
        
        return dim_dict
    
    def __init_extent(self):
        
        extent_dict = {"xmin" : [], "xmax": [], "ymin": [], "ymax": []}
        
        with rasterio.open(self.source) as raster:
            
            extent_dict["xmin"].append(raster.bounds[0])
            extent_dict["xmax"].append(raster.bounds[2])
            extent_dict["ymin"].append(raster.bounds[1])
            extent_dict["ymax"].append(raster.bounds[3])
            
        return extent_dict
    
    def __init_resolution(self):
        
        x_num = self.dimensions["width"][0]
        y_num = self.dimensions["height"][0]
        
        x_range = self.extent["xmax"][0] - self.extent["ymin"][0]
        y_range = self.extent["ymax"][0] - self.extent["ymin"][0]
        
        x_res = x_range / x_num
        y_res = y_range / y_num
        
        res_dict = {"x": x_res, "y": y_res}
        
        return res_dict
    
    def __init_crs(self):
        
        with rasterio.open(self.source) as raster:
            
            crs = raster.crs
            
        return crs
    
        




import pysyncrosim as ps
from pysyncrosim.environment import _environment
import os
import io
import pandas as pd
import numpy as np

class Scenario(object):
    """
    A class representing a SyncroSim Scenario.
    
    """
    __datasheets = None
    __results = None
    
    def __init__(self, sid=None, name=None, project=None, library=None):
        self.__sid = sid
        self.__name = name
        self.__project = project
        self.__library = library
        
        # Initialize when in a SyncroSim environment
        if sid is None and name is None and project is None and \
            library is None:
            e = _environment()
            self.__library = ps.library(e.library_filepath.item())
            self.__project = self.__library.projects(
                pid=e.project_id.item())
            self.__sid = int(e.scenario_id.item())
            temp_df = self.__library.scenarios()
            self.__name = temp_df[
                temp_df["ScenarioID"] == self.__sid]["Name"].item()
            # revisit these!!
            self.__env = e.transfer_directory.item()
            self.__temp = e.temp_directory
        else:
            self.__env = None
        self.__owner = None
        self.__date_modified = None
        self.__readonly = None
        self.__project_id = None
        self.__info = None
        # All None attributes assigned in __init_info()
        self.__init_info()
        self.__description = self.__init_description()
        self.__is_result = self.__init_is_result()
        self.__parent_id = self.__init_parent_id()
        self.__dependencies = self.__init_dependencies()
        
    @property
    def sid(self):
        """
        Retrieves Scenario ID.

        Returns
        -------
        Int
            Scenario ID.

        """
        return self.__sid
    
    @property
    def name(self):
        """
        Retrieves or sets a Scenario name.
        
        Parameters
        ----------
        Value : String
            New Scenario name.

        Returns
        -------
        String
            Scenario name.

        """
        return self.__name
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--name=%s" % value, "--sid=%d" % self.sid]
        self.library.session._Session__call_console(args)
        # Reset information
        self.__init_info()
        
    @property
    def project(self):
        """
        Retrieves Project class instance associated with this Scenario.

        Returns
        -------
        Project
            SyncroSim Project class instance.

        """
        return self.__project
    
    @property
    def library(self):
        """
        Retrieves Library class instance associated with this Library.

        Returns
        -------
        Library
            SyncroSim Library class instance.

        """
        return self.__library
    
    @property
    def info(self):
        """
        Retrieves Scenario information.

        Returns
        -------
        pandas DataFrame
            Scenario information.

        """
        return self.__info
    
    @property
    def owner(self):
        """
        Gets or sets the owner of this Scenario.

        Returns
        -------
        String
            Owner of this Scenario.

        """
        return self.__owner
    
    @owner.setter
    def owner(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--owner=%s" % value, "--sid=%d" % self.sid]
        self.library.session._Session__call_console(args)
        # Reset information
        self.__init_info()
        
    @property
    def date_modified(self):
        """
        Gets the last date this Scenario was modified.

        Returns
        -------
        String
            Last date modified.

        """
        return self.__date_modified
    
    @property
    def readonly(self):
        """
        Gets or sets the read-only status for this Scenario.

        Returns
        -------
        String
            "yes" if this Scenario is read-only, "no" otherwise.

        """
        return self.__readonly
    
    @readonly.setter
    def readonly(self, value):
        if value is True or value == "yes":
            ro = "yes"
        elif value is False or value == "no":
            ro = "no"
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--readonly=%s" % ro, "--sid=%d" % self.sid]
        self.library.session._Session__call_console(args)
        # Reset information
        self.__init_info()
        
    @property
    def project_id(self):
        """
        Gets the Project ID associated with this Scenario.

        Returns
        -------
        Integer
            Project ID.

        """
        return self.__project_id
    
    @property
    def description(self):
        """
        Gets or sets the Scenario description.

        Returns
        -------
        String
            Scenario description.

        """
        return self.__description
    
    @description.setter
    def description(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location,
                "--description=%s" % value, "--sid=%d" % self.sid]
        self.library.session._Session__call_console(args)
        self.__description = self.__init_description()
    
    @property
    def is_result(self):
        """
        Retrieves information about whether this Scenario is a Results
        Scenario or not.

        Returns
        -------
        Logical
            Whether Scenario is a Results Scenario.

        """
        return self.__is_result
    
    @property
    def parent_id(self):
        """
        Retrieves the ID of the parent Scenario of a Results Scenario.

        Returns
        -------
        Int
            Parent ID of a Results Scenario.

        """
        return self.__parent_id
    
    def datasheets(self, name=None, summary=True, optional=False, empty=False,
                   filter_column=None, filter_value=None, include_key=False):
        """
        Retrieves a DataFrame of Scenario Datasheets.
        
        Parameters
        ----------
        name : String, optional
            Datasheet name. The default is None.
        summary : Logical or String, optional
            Whether to list package Datasheets or core Datasheets. The default
            is True.
        optional : Logical, optional
            Return optional columns. The default is False.
        empty : Logical, optional
            If True, returns an empty Datasheet. The default is False.
        filter_column : String
            The column and value to filter the output Datasheet by 
            (e.g. "TransitionGroupID=20"). The default is None.
        include_key : Logical, optional
            Whether to include the primary key of the Datasheet, corresponding
            to the SQL database. Default is False.

        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Datasheet 
            information including Package, Name, and Display Name.
            If `optional=True`, also returns Scope, Is Single, and Is Output.

        """
        
        self.__datasheets = self.library.datasheets(name, summary, optional,
                                                    empty, "Scenario",
                                                    filter_column, 
                                                    filter_value, include_key,
                                                    self.sid)
        return self.__datasheets
    

    def datasheet_rasters(self, datasheet, column=None, iteration=None,
                         timestep=None, filter_column=None, filter_value=None,
                         path_only=False):
        """
        Retrieves spatial data columns from one or more SyncroSim Datasheets.

        Parameters
        ----------
        datasheet : String
            The name of a SyncroSim Datasheet containing raster data.
        column : String
            The column in the Datasheet containing the raster data. If no 
            column selected, then datasheet_rasters will attempt to find one.
        iteration : Int, List, or Range, optional
            The iteration to subset by. The default is None.
        timestep : Int, List, or Range, optional
            The timestep to subset by. The default is None.
        filter_column : String
            The column to filter the output rasters by 
            (e.g. "TransitionGroupID=20"). The default is None.
        filter_value : String, Int, Logical
            The value to filter the filter_column by. The default is None.
        path_only : Logical
            Instead of returning a Raster Class Instance, a filepath to the
            raster is returned. The default is False.

        Returns
        -------
        Raster or List of Rasters 
            Raster class instance or List of these.

        """
        # Type Checks
        if not isinstance(datasheet, str):
            raise TypeError("datasheet must be a String")
            
        if column is not None and not isinstance(column, str):
            raise TypeError("column must be a String")
            
        if iteration is not None and not isinstance(iteration, int)\
            and not isinstance(iteration, np.int64)\
                and not isinstance(iteration, list)\
                    and not isinstance(iteration, range):
            raise TypeError("iteration must be an Integer, List, or Range")
            
        if timestep is not None and not isinstance(timestep, int)\
            and not isinstance(timestep, np.int64)\
                and not isinstance(timestep, list)\
                    and not isinstance(timestep, range):
            raise TypeError("timestep must be an Integer, List, or Range")
        
        # Check that self is Results Scenario
        if self.is_result == "No":
            raise ValueError("Scenario must be a Results Scenario")
            
        # Check that Datasheet has package prefix
        datasheet = self.library._Library__check_datasheet_name(datasheet)
        
        # Retrieve Datasheet as DataFrame
        d = self.datasheets(name = datasheet, filter_column = filter_column,
                            filter_value = filter_value)
        
        # Check if column is raster column
        args = ["--list", "--columns", "--allprops",
                "--sheet=%s" % datasheet, "--csv", 
                "--lib=%s" % self.library.location]
        props = self.library.session._Session__call_console(args, decode=True,
                                                            csv=True)
        props = pd.read_csv(io.StringIO(props))
        props["is_raster"] = props.Properties.str.contains(r"isRaster\^True")
        
        if (props.is_raster == False).all():
            raise ValueError(
                f"No raster columns found in Datasheet {datasheet}")
          
        # If no raster column specified, find the raster column
        if column is None:
            if len(props[props.is_raster == True]) > 1:
                raise ValueError(
                    "> 1 raster output column available, please specify.")
            column = props[props.is_raster == True].Name.values[0]
            
        if not (props.Name == column).any():
            raise ValueError(
               f"Column {column} not found in Datasheet {datasheet}")
            
        col_props = props[props.Name == column]
        
        if col_props.is_raster is False:
            raise ValueError(f"Column {column} is not a raster column")
            
        # TODO: Get band column if it exists
        # if col_props.Properties.str.contains("bandColumn").any():
        #     prop_split = col_props.Properties.str.split("!").values[0]
        #     band_column = [b for b in prop_split if b.startswith("bandColumn")]
        #     col_props["band_column"] = band_column[0].split("^")[1]
        
        if isinstance(iteration, range):
            iteration = list(iteration)
        
        if iteration is not None:
            if isinstance(iteration, int):
                
                if iteration > d.Iteration.max():
                    raise ValueError(
                        "Specified iteration above range of plausible values")
                elif iteration <= 0:
                    raise ValueError("iteration cannot be below 1")
                    
                d = d.loc[d["Iteration"] == iteration]
                
            if isinstance(iteration, list):
                
                if any(x > d.Iteration.max() for x in iteration) or any(
                        x < 1 for x in iteration):
                    raise ValueError("Some iteration values outside of range")
                    
                d = d.loc[d["Iteration"].isin(iteration)]
                
            d = d.reset_index()
            
        if isinstance(timestep, range):
            timestep = list(timestep)
            
        if timestep is not None:
            if isinstance(timestep, int):
                
                if timestep > d.Timestep.max():
                    raise ValueError(
                        "Specified timestep above range of plausible values")
                if timestep < d.Timestep.min():
                    raise ValueError(
                        "Specified timestep below range of plausible values")
                    
                d = d.loc[d["Timestep"] == timestep]
                
            if isinstance(timestep, list):
                
                if any(x > d.Timestep.max() for x in timestep) or any(
                        x < d.Timestep.min() for x in timestep):
                    raise ValueError("Some timestep values outside of range")
                
                d = d.loc[d["Timestep"].isin(timestep)]
                
            d = d.reset_index()
                
        # Create empty list to store raster objects
        raster_list = []
        
        # Search for the following raster tifs in all possible output places
        raster_tifs = d[column].values
        rpaths = []
        
        # Find folder with raster data - search in input, temp, and output
        if self.__env is None:
            
            for folder in [".input", ".temp", ".output"]:
                
                if folder != ".temp":
                    lib_dir = self.__find_output_fpath(
                        self.library.location + folder, datasheet)
                else:
                    lib_dir = self.library.location + folder
                
                for raster_tif in raster_tifs:
                    for root, dirs, files in os.walk(lib_dir):
                        if raster_tif in files:
                            rpaths.append(os.path.join(root, raster_tif))
                        else:
                            break
                        
                    if len(rpaths) == 0:
                        break
                    
                if len(rpaths) !=0:
                    break
                
        else:
            e = _environment()
            f_base_path = e.input_directory.item()
            fpath = self.__find_output_fpath(f_base_path, datasheet)
            
            for i in range(0, len(d)):
  
                # Index column with raster data
                rpaths = os.path.join(fpath, d[column].loc[i])
        
        # Return only filepaths to rasters if path_only is True
        if path_only:
            return rpaths

        
        # Iterate through all raster files in Datasheet
        for i in range(0, len(rpaths)):
            
            # Open and append each raster from the Datasheet
            if "Iteration" in d.columns:
                if "Timestep" in d.columns:
                    raster = ps.Raster(rpaths[i], iteration = d["Iteration"].loc[i],
                                       timestep = d["Timestep"].loc[i])
                else:
                    raster = ps.Raster(rpaths[i], iteration = d["Iteration"].loc[i])
            elif "Timestep" in d.columns:
                raster = ps.Raster(rpaths[i], timestep=d["Timestep"].loc[i])
            else:
                raster = ps.Raster(rpaths[i])

            raster_list.append(raster)
            
        if len(raster_list) == 1:
            return raster_list[0]
        else:
            return raster_list
    
    def save_datasheet(self, name, data):
        """
        Saves a pandas DataFrame as a SyncroSim Datasheet.

        Parameters
        ----------
        name : String
            Name of the SyncroSim Datasheet.
        data : pandas.DataFrame
            Data to be saved to SyncroSim Datasheet.

        Returns
        -------
        None.

        """
        self.library.save_datasheet(name, data, "Scenario", self.sid)
    
    def delete(self, force=False):
        """
        Deletes a Scenario.

        Parameters
        ----------
        force : Logical, optional
            If True, does not ask the user for permission to delete the 
            Scenario. The default is False.

        Returns
        -------
        None.

        """
        
        self.library.delete(project=self.project, scenario=self, force=force)
    
    def copy(self, name=None):
        """
        Creates a copy of an existing Scenario class instance.

        Parameters
        ----------
        name : String
            Name of the new Scenario. If no name is provided, the copied 
            Scenario is named after the existing Scenario. The default is None.

        Returns
        -------
        Scenario
            SyncroSim Scenario class instance.

        """
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        
        if name is None:
            name = self.name + " - Copy"
        
        args = ["--copy", "--scenario", "--slib=%s" % self.library.name,
                "--sid=%d" % self.sid, "--name=%s" % name]
        
        self.library.session._Session__call_console(args)
        
        # Reset Scenarios
        self.library._Library__scenarios = None
        self.library._Library__init_scenarios()
        s = self.library._Library__get_scenario(name=name)
        
        return ps.Scenario(s["ScenarioID"].values[0],
                           s["Name"].values[0], self.project, self.library)
    
    def dependencies(self, dependency=None, remove=False, force=False):
        """
        Gets, sets, or removes dependencies from a Scenario.

        Parameters
        ----------
        dependency : Scenario, Int, String, or List, optional
            Scenario(s) to be added or removed as dependencies. If None, then 
            returns a DataFrame of dependencies. The default is None.
        remove : Logical, optional
            If False, adds the dependency. If True, removes the dependency. 
            The default is False.
        force : Logical, optional
            If True, then does not prompt the user when removing a dependency.
            The default is False.

        Raises
        ------
        TypeError
            If arguments are not of the correct Type, throws an error.

        Returns
        -------
        pandas.DataFrame
            If no dependencies are specified, then returns a pandas DataFrame
            of existing dependencies for the given Scenario.

        """
        
        if dependency is None:
            return self.__dependencies
        
        if not isinstance(dependency, list):
            dependency = [dependency]
        
        for d in dependency:
            # Check types
            if isinstance(d, ps.Scenario):
                d_name = d.name
                d = d.sid
                
            elif isinstance(d, str):
                # check if scenario exists
                d_name = d
                d = self.library._Library__scenarios[
                    self.library._Library__scenarios.Name==d]
                
                if len(d) > 1:
                    raise ValueError(
                        "dependency name not unique, use ID or Scenario")
                else:
                    d = d["ScenarioID"].item()
                
            elif isinstance(d, int) or isinstance(d, np.int64):
                # check if scenarios exists
                d_name = self.library._Library__scenarios[
                         self.library._Library__scenarios["ScenarioID"]==d
                         ].Name.item()
                
            else:
                raise TypeError(
                    "dependency must be a Scenario, String, Integer, or List")
            
            # Add dependency
            if remove is False:
                # Remove and re-add to guarantee order
                if d in self.__dependencies.ID.values:
                    args = ["--delete", "--dependency", 
                            "--lib=%s" % self.library.location,
                            "--sid=%d" % self.sid, "--did=%d" % d, "--force"]
                    self.library.session._Session__call_console(args)
                
                # Re-add now
                args = ["--create", "--dependency",
                        "--lib=%s" % self.library.location,
                        "--sid=%d" % self.sid, "--did=%d" % d]
                
                try:
                    self.library.session._Session__call_console(args)
                    # Reset dependencies
                    self.__dependencies = self.__init_dependencies()
                    
                except RuntimeError as e:
                    print(e)
                
            # Remove dependency
            elif remove is True:
                args = ["--delete", "--dependency",
                        "--lib=%s" % self.library.location,
                        "--sid=%d" % self.sid, "--did=%d" % d]
                
                if force is False:
                    answer = input (
                        f"Do you really want to remove dependency {d_name} \
                            (Y/N)?")
                elif force is True:
                    answer = "Y"
                else: 
                    raise TypeError("force must be a Logical")
                    
                if answer == "Y":
                    args += ["--force"]
                else:
                    print(f"dependency {d_name} not removed")
                    return
                try:
                    self.library.session._Session__call_console(args)
                    # Reset dependencies
                    self.__dependencies = self.__init_dependencies()
                except RuntimeError as e:
                    print(e)
                         
            else: 
                raise TypeError("remove must be a Logical")
    
    def ignore_dependencies(self, value=None):
        """
        Retrieves or sets the Datafeeds to ignore for a Scenario.

        Parameters
        ----------
        value : String, optional
            Datafeeds to ignore. If more than one, separate with a comma, but 
            keep contained within the same String. If None, then a String of 
            Datafeeds to ignore is returned. The default is None.

        Returns
        -------
        String
            List of Datafeeds to ignore.

        """
        
        # Set datafeeds to ignore from a Scenario's dependencies
        if value is None:
            scn_info = self.library._Library__scenarios
            return scn_info[
                scn_info[
                    "ScenarioID"] == self.sid]["IgnoreDependencies"].item()
        elif isinstance(value, str):
            args = ["--setprop", "--lib=%s" % self.library.location,
                    "--ignoredeps='%s'" % value, "--sid=%d" % self.sid]
            self.library.session._Session__call_console(args)
            
            # Reset Scenario information
            self.library._Library__init_scenarios()
        else:
            raise TypeError("value must be a String")
    
    def merge_dependencies(self, value=None):
        """
        Retrieves or sets whether or not a Scenario is configured to merge
        dependencies at run time.

        Parameters
        ----------
        value : Logical, optional
            If True, the Scenario will be configured to merge dependencies at
            run time. If None, then the current configuration of merge 
            dependencies is returned. The default is None.

        Returns
        -------
        String
            Whether or not the Scenario is configured to merge dependencies at 
            run time.

        """
        
        scn_info = self.library._Library__scenarios
        merge_dep_status =  scn_info[scn_info[
                "ScenarioID"] == self.sid]["MergeDependencies"].item()
        
        if value is None:
            return merge_dep_status
        
        elif value is True:
            if merge_dep_status == "yes":
                print("Scenarios already configured to merge at runtime")
                return
            else:
                merge_deps = "yes"
            
        elif value is False:
            if merge_dep_status == "no":
                print("Scenarios already configured to not merge at runtime")
                return
            else:
                merge_deps = "no"
            
        else:
            raise TypeError("value must be a Logical")
            
        args = ["--setprop", "--lib=%s" % self.library.location,
                "--mergedeps=%s" % merge_deps, "--sid=%d" % self.sid]
        self.library.session._Session__call_console(args)
        
        # Reset Scenario information
        self.library._Library__init_scenarios()
        
    def run(self, jobs=1, copy_external_inputs=False):
        """
        Runs a Scenario.

        Parameters
        ----------
        jobs : Int, optional
            Number of multiprocessors to use when running a Scenario. The 
            default is 1.
        copy_external_inputs : Logical, optional
            If False, then a copy of external input files (e.g. GeoTIFF files)
            is not created for each job. Otherwise, a copy of external inputs 
            is created for each job. Applies only when jobs > 1. The default is
            False.

        Returns
        -------
        Scenario
            SyncroSim Scenario class instance.

        """
        if not isinstance(jobs, int) and not isinstance(jobs, np.int64):
            raise TypeError("jobs must be an Integer")
        
        # Runs the scenario
        args = ["--run", "--lib=%s" % self.library.location,
                "--sid=%d" % self.__sid, "--jobs=%d" % jobs]
        
        if jobs > 1 and copy_external_inputs is False:
            args += ["--noextfiles"]
            
        self.library.session._Session__call_console(args)
        
        # Reset Project Scenarios
        self.project._Project__scenarios = None
        
        # Have this print a statement based on output message of whether the 
        # run was successful or not

        # Reset results
        self.__results = None
        
        # Retrieve Results Scenario ID
        # Also resets scenarios and results info
        result_id = self.results()["ScenarioID"].values[-1]
        
        # Return Results Scenario
        result_scn = self.library.scenarios(project=self.project,
                                            name=None,
                                            sid=result_id)
        
        return result_scn
    
    def run_log(self):
        """
        Returns a run log for a Results Scenario.

        Returns
        -------
        pandas.DataFrame
            Information from the SyncroSim run of a Results Scenario.

        """
        try:
            args = ["--list", "--runlog", "--lib=%s" % self.library.name,
                    "--sid=%d" % self.sid]
            return self.library._Library__console_to_csv(args)
        except RuntimeError as e:
            print(e)
        
    def results(self, sid=None):
        """
        Retrieves DataFrame of Results Scenarios or retrieves a Results 
        Scenario instance for this Scenario.
        
        Parameters
        ----------
        sid : Int, optional
            Scenario ID of Results Scenario. Default is None.

        Returns
        -------
        pandas.DataFrame or Scenario
            DataFrame of Results Scenarios, including information about 
            Scenario ID, Project ID, Name, and Is Result, or SyncroSim
            Scenario class instance.

        """
        # Type checks
        if sid is not None and not isinstance(
                sid, int) and not isinstance(sid, np.int64):
            raise TypeError("Scenario ID must be an Integer")
        
        # Retrieve specified Results Scenario for this scenario
        if sid is not None:
            scn = self.library.scenarios(project=self.project,
                                         name=None,
                                         sid=sid)
            if scn.is_result == "Yes":
                return scn
            else:
                raise ValueError(f'Scenario [{sid}] is not a Results Scenario')
        
        elif self.__results is None:
            s = self.project.scenarios()
            pat = '['+str(self.__sid)+']'
            self.__results = s[s.Name.str.contains(pat)]
            return self.__results
        else:
            return self.__results
        
    def __init_info(self):
        # Set Scenario information
        scn_info = self.library.scenarios(project=self.project.pid,
                                          optional=True)
        scn_info = scn_info[scn_info["ScenarioID"] == self.sid]
        self.__owner = scn_info["Owner"].item()
        self.__date_modified = scn_info["DateLastModified"].item()
        self.__readonly = scn_info["IsReadOnly"].item()
        self.__project_id = scn_info["ProjectID"].item()
        self.__info = scn_info.set_axis(
            ["Value"], axis=0, inplace=False
            ).T.rename_axis("Property").reset_index()
        
    def __init_description(self):
        args = ["--list", "--description", "--lib=%s" % self.library.location,
                "--sid=%d" % self.sid]
        return self.library.session._Session__call_console(args, decode=True)
    
    def __init_is_result(self):
        
        # Find out if result scenario
        scn_info = self.library._Library__scenarios
        scn_info = scn_info[scn_info["ScenarioID"] == self.sid]
        return scn_info["IsResult"].values[0]
    
    def __init_parent_id(self):
        
        # Find out parent ID if result scenario
        scn_info = self.library._Library__scenarios
        scn_info = scn_info[scn_info["ScenarioID"] == self.sid]
        parent_id = scn_info["ParentID"].values[0]
        if type(parent_id) == float:
            return int(parent_id)
        else:
            return parent_id
    
    def __init_dependencies(self):
        
        # Find all dependencies for a Scenario
        args = ["--list", "--dependencies", "--lib=%s" % self.library.location,
                "--sid=%d" % self.sid, ]
        return self.library._Library__console_to_csv(args) 
    
    def __find_output_fpath(self, f_base_path, datasheet):
                
        # If package is not included in name, add it
        if datasheet.startswith(self.library.package) is False:
            if datasheet.startswith("core") is False:
                datasheet = self.library.package + "_" + datasheet

        fpath = os.path.join(f_base_path, f"Scenario-{self.sid}", datasheet)
        
        return fpath


import pysyncrosim as ps

class Project(object):
    """
    A class to represent a SyncroSim Project.
        
    """
    __scenarios = None
    __datasheets = None
    
    def __init__(self, pid, name, library):
        """
        Initializes a pysyncrosim Project instance.

        Parameters
        ----------
        pid : Int
            Project ID.
        name : String
            Project name.
        library : Library
            pysyncrosim Library instance.

        Returns
        -------
        None.

        """
        self.__pid = pid
        self.__name = name
        self.__library = library
        self.__owner = None
        self.__date_modified = None
        self.__readonly = None
        self.__info = None
        # All None attributes created by __init_info()
        self.__init_info()
        self.__description = self.__init_description()
        
    @property
    def pid(self):
        """
        Retrieves Project ID.

        Returns
        -------
        Int
            Project ID.

        """
        return self.__pid
    
    @property
    def name(self):
        """
        Retrieves or sets Project name.

        Returns
        -------
        String
            Project name.

        """
        return self.__name 
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--name=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
    
    @property
    def library(self):
        """
        Retrieves the Library the Project belongs to.

        Returns
        -------
        Library
            SyncroSim Library class instance.

        """
        return self.__library
    
    @property
    def info(self):
        """
        Retrieves Project information.

        Returns
        -------
        pandas DataFrame
            Project information.

        """
        return self.__info
    
    @property
    def owner(self):
        """
        Gets or sets the owner of this Project.

        Returns
        -------
        String
            Owner of this Project.

        """
        return self.__owner
    
    @owner.setter
    def owner(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--owner=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
        
    @property
    def date_modified(self):
        """
        Gets the last date this Project was modified.

        Returns
        -------
        String
            Last date modified.

        """
        return self.__date_modified
    
    @property
    def readonly(self):
        """
        Gets or sets the read-only status for this Project.

        Returns
        -------
        String
            "yes" if this Project is read-only, "no" otherwise.

        """
        return self.__readonly
    
    @readonly.setter
    def readonly(self, value):
        if value is True or value == "yes":
            ro = "yes"
        elif value is False or value == "no":
            ro = "no"
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.library.location, 
                "--readonly=%s" % ro, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__init_info()
    
    @property
    def description(self):
        """
        Gets or sets the Project description.

        Returns
        -------
        String
            Project description.

        """
        return self.__description
    
    @description.setter
    def description(self, value):
        args = ["--setprop", "--lib=%s" % self.library.location,
                "--description=%s" % value, "--pid=%d" % self.pid]
        self.library.session._Session__call_console(args)
        self.__description = None
        self.__description = self.__init_description()

    def scenarios(self, name=None, sid=None, optional=False, summary=None,
                  results=False):
        """
        Retrieve a DataFrame of Scenarios in this Project.
        
        Parameters
        ----------
        name : String, optional
            Scenario name. The default is None.
        sid : Int, optional
            Scenario ID. The default is None.
        optional : Logical, optional
            Return optional information. The default is False.
        summary : Logical, optional
            If set to False, then returns all Scenarios as SyncroSim
            Scenario instances. The default is None.

        results : Logical, optional
            Return only a list of Results Scenarios. The default is False.
        
        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Scenario 
            information including Scenario ID, Project ID, Name, and Is Result.
            If `optional=True`, also returns Parent ID, Owner, Last Modified, 
            Read Only, Merge Dependencies, Ignore Dependencies, and Auto Gen
            Tags.

        """
        
        self.__scenarios = self.library.scenarios(name=name, project=self, 
                                                  sid=sid,
                                                  pid=self.pid,
                                                  optional=optional,
                                                  summary=summary,
                                                  results=results)
        
        return self.__scenarios

    def datasheets(self, name=None, summary=True, optional=False, empty=False,
                   filter_column=None, filter_value=None, include_key=False):
        """
        Retrieves a DataFrame of Project Datasheets.
        
        Parameters
        ----------
        name : String, optional
            Datasheet name. The default is None.
        summary : Logical or String, optional
            Whether to list package Datasheets or core Datasheets. The default
            is True.
        optional : Logical, optional
            Return optional columns. The default is False.
        empty : Logical, optional
            If True, returns an empty Datasheet. The default is False.
        filter_column : String
            The column and value to filter the output Datasheet by 
            (e.g. "TransitionGroupID=20"). The default is None.
        include_key : Logical, optional
            Whether to include the primary key of the Datasheet, corresponding
            to the SQL database. Default is False.

        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Datasheet 
            information including Package, Name, and Display Name.
            If `optional=True`, also returns Scope, Is Single, and Is Output.

        """
        
        self.__datasheets = self.library.datasheets(name, summary, optional,
                                                    empty, "Project",
                                                    filter_column, 
                                                    filter_value, include_key,
                                                    self.pid)
        return self.__datasheets
    
    def delete(self, scenario=None, force=False):
        """
        Deletes a Project or Scenario.

        Parameters
        ----------
        scenario : Scenario, String, or Int, optional
            Scenario to delete. The default is None.
        force : Logical, optional
            If True, does not prompt the user to confirm deletion. The default
            is False.

        Returns
        -------
        None.

        """
        
        self.library.delete(project=self, scenario=scenario, force=force)

    def save_datasheet(self, name, data):
        """
        Saves a Project-scoped Datasheet.

        Parameters
        ----------
        name : String
            Name of Datasheet to save.
        data : pandas.DataFrame
            DataFrame to save as Datasheet.

        Returns
        -------
        None.

        """
        
        self.library.save_datasheet(name, data, "Project", self.pid)
    
    def copy(self, name=None):
        """
        Creates a copy of an existing Project.

        Parameters
        ----------
        name : String, optional
            Name of the new Project. If no name is given, the copied 
            Project is named after the source Project. The default is None.

        Returns
        -------
        Project
            SyncroSim Project class instance.

        """
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        
        if name is None:
            name = self.name + " - Copy"
        
        args = ["--copy", "--project", "--slib=%s" % self.library.name,
                "--pid=%d" % self.pid, "--name=%s" % name, "--copychildren"]
        
        self.library.session._Session__call_console(args)
        
        # Reset Projects
        self.library._Library__projects = None
        self.library._Library__init_projects()
        p = self.library._Library__get_project(name=name)
        
        return ps.Project(p["ProjectID"].values[0],
                          p["Name"].values[0], self.library)
    
    def __init_info(self):
        # Set projects
        self.library.projects()
        proj_info = self.library._Library__projects
        proj_info = proj_info[proj_info["ProjectID"] == self.pid]
        self.__owner = proj_info["Owner"].item()
        self.__date_modified = proj_info["DateLastModified"].item()
        self.__readonly = proj_info["IsReadOnly"].item()
        self.__info = proj_info.set_axis(
            ["Value"], axis=0, inplace=False
            ).T.rename_axis("Property").reset_index()
        
    def __init_description(self):
        args = ["--list", "--description",
                "--lib=%s" % self.library.location,
                "--pid=%d" % self.pid]
        return self.library.session._Session__call_console(
            args, decode=True)





import pysyncrosim as ps
import pytest
import pandas as pd
import math
import numpy as np
import os
import rasterio
# import re

def test_session_attributes():
    
    mySession = ps.Session()
    
    # Test init
    assert isinstance(mySession, ps.Session)
    
    with pytest.raises(ValueError, match="The location is not valid"):
        mySession = ps.Session(location="bad/location")
        
    # Test version method
    assert isinstance(mySession.version(), str)
    assert "Version is:" in mySession.version()
    
    # Test packages method
    assert isinstance(mySession.packages(), pd.DataFrame)
    assert isinstance(mySession.packages(installed=False), pd.DataFrame)
    assert isinstance(mySession.packages(installed="BASE"), pd.DataFrame)
    
    with pytest.raises(TypeError,
                       match="installed must be Logical or 'BASE'"):
        mySession.packages(installed=1)
       
        
def test_session_package_functions():
    
    mySession = ps.Session()
    
    # Test add_packages, remove_packages, update_packages methods
    with pytest.raises(TypeError, match="packages must be a String or List"):
        mySession.add_packages(1)
    with pytest.raises(TypeError, match="packages must be a String or List"):
        mySession.remove_packages(1)
    with pytest.raises(TypeError, match="packages must be a String or List"):
        mySession.update_packages(1)
        
    with pytest.raises(TypeError, match="all packages must be Strings"):
        mySession.add_packages(["helloworldSpatial", 1])
    with pytest.raises(TypeError, match="all packages must be Strings"):
        mySession.remove_packages(["helloworldSpatial", 1])
    with pytest.raises(TypeError, match="all packages must be Strings"):
        mySession.update_packages(["helloworldSpatial", 1])
        
    mySession.add_packages("helloworldSpatial")
    assert "helloworldSpatial" in mySession.packages()["Name"].values
    
    mySession.remove_packages("helloworldSpatial")
    assert "helloworldSpatial" not in mySession.packages()["Name"].values
    
    mySession.add_packages("helloworldSpatial")
    
def test_helper():
    
    mySession = ps.Session()
    mySession.add_packages("stsim")
    
    # Type checking
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument"):
        ps.library()
        
    with pytest.raises(TypeError, match="name must be a String"):
        ps.library(name=1)
        
    with pytest.raises(
            TypeError,
            match="session must be None or pysyncrosim Session instance"):
        ps.library(name="Test", session=1)
        
    with pytest.raises(TypeError, match="package must be a String"):
        ps.library(name="Test", package=1)
        
    with pytest.raises(TypeError, match="addons must be a String"):
        ps.library(name="Test", addons=1)
        
    with pytest.raises(TypeError, match="templates must be a String"):
        ps.library(name="Test", template=1)
        
    with pytest.raises(TypeError, match="forceUpdate must be a Logical"):
        ps.library(name="Test", forceUpdate="True")
        
    with pytest.raises(TypeError, match="overwrite must be a Logical"):
        ps.library(name="Test", overwrite="False")

    # Test package installation
    mySession.remove_packages("stsim")
    with pytest.raises(ValueError, match="The package stsim is not installed"):
        ps.library(name="Test", package="stsim", session=mySession)
    mySession.add_packages("stsim")
        
    # Test Library path
    with pytest.raises(ValueError, match="Path to Library does not exist"):
        ps.library("path/to/library")
        
    # Test template
    with pytest.raises(ValueError,
                       match="Template test does not exist in package"):
        ps.library("Test", template="test")
      
    # Test output
    myLibrary = ps.library(name="Test")
    assert isinstance(myLibrary, ps.Library)
    
def test_library_attributes():
    
    myLibrary = ps.library(name="Test", overwrite=True)
    
    # Check attributes
    assert isinstance(myLibrary.name, str)
    assert isinstance(myLibrary.session, ps.Session)
    assert isinstance(myLibrary.location, str)
    assert os.path.isfile(myLibrary.location)
    assert isinstance(myLibrary.package, str)
    assert isinstance(myLibrary.addons, pd.DataFrame)

def test_library_projects():
    
    myLibrary = ps.library(name="Test", overwrite=True)
    
    # Test inputs
    with pytest.raises(TypeError, match="name must be a String"):
        myLibrary.projects(name=1)
        
    with pytest.raises(TypeError, match="pid must be an Integer"):
        myLibrary.projects(pid="1")
        
    with pytest.raises(TypeError, match="summary must be a Logical"):
        myLibrary.projects(summary="False")
        
    with pytest.raises(TypeError, match="overwrite must be a Logical"):
        myLibrary.projects(overwrite="False")
        
    with pytest.raises(ValueError, match="Project ID 2 does not exist"):
        myLibrary.projects(pid=2)
    
    with pytest.raises(ValueError,
                       match="Project ID 1 does not match Project name test2"):
        myLibrary.projects(name="test")
        myLibrary.projects(name="test2", pid=1)
        
    # Test outputs
    assert isinstance(myLibrary.projects(), pd.DataFrame)
    assert isinstance(myLibrary.projects(name="test"), ps.Project)  
    assert isinstance(myLibrary.projects(summary=False), list)
        
def test_library_scenarios():
    
    myLibrary = ps.library(name="Test", overwrite=True)
    myLibrary.projects(name="test")
    
    with pytest.raises(
            TypeError, 
            match="name must be a String, Integer, or List of these"):
        myLibrary.scenarios(name=pd.DataFrame())
        
    with pytest.raises(
            TypeError,
            match="project must be Project instance, String, or Integer"):
        myLibrary.scenarios(project=[1])
        
    with pytest.raises(TypeError, match="sid must be an Integer"):
        myLibrary.scenarios(sid="1")
    
    with pytest.raises(TypeError, match="pid must be an Integer"):
        myLibrary.scenarios(pid="1")
    
    with pytest.raises(TypeError, match="overwrite must be a Logical"):
        myLibrary.scenarios(overwrite="False")
        
    with pytest.raises(TypeError, match="optional must be a Logical"):
        myLibrary.scenarios(optional=1)
        
    with pytest.raises(TypeError, match="summary must be a Logical"):
        myLibrary.scenarios(summary="True")
        
    with pytest.raises(ValueError, match="Scenario ID 2 does not exist"):
        myLibrary.scenarios(sid=2)
        
    # Test scenarios method outputs
    assert myLibrary.scenarios(name="test").name == "test"
    assert myLibrary.scenarios(name="test2", sid=1).name == "test"
    assert isinstance(myLibrary.scenarios(name="test"), ps.Scenario)
    assert isinstance(myLibrary.scenarios(name="test", sid=1), ps.Scenario)
    assert isinstance(myLibrary.scenarios(), pd.DataFrame)
    
    myLibrary.projects(name="project2")
    with pytest.raises(ValueError, match="More than one Project in Library"):
        myLibrary.scenarios(summary=False)
    
    assert isinstance(myLibrary.scenarios(), pd.DataFrame) 
    assert isinstance(myLibrary.scenarios(pid=1, summary=False), ps.Scenario)
    assert len(myLibrary.scenarios(pid=1).columns) == 4
    assert len(myLibrary.scenarios(pid=1, optional=True).columns) == 11
    assert myLibrary.scenarios(name="test", overwrite=True).sid != 1
    assert all(myLibrary.scenarios(project=1) == myLibrary.scenarios(pid=1))
    assert all(
        myLibrary.scenarios(project="test") == myLibrary.scenarios(pid=1))
    myProject = myLibrary.projects("test")
    assert all(myLibrary.scenarios(
        project=1) == myLibrary.scenarios(project=myProject))
    
def test_library_datasheets():
    
    myLibrary = ps.library(name="Test", overwrite=True)
    
    # Test datasheets method inputs
    with pytest.raises(TypeError, match="name must be a String"):
        myLibrary.datasheets(name=1)
        
    with pytest.raises(TypeError, match="summary must be a Logical or 'CORE'"):
        myLibrary.datasheets(summary=1)
        
    with pytest.raises(TypeError, match="optional must be a Logical"):
        myLibrary.datasheets(optional=[1, 2, 3])
        
    with pytest.raises(TypeError, match="filter_column must be a String"):
        myLibrary.datasheets(filter_column=1)
    
    with pytest.raises(
            RuntimeError,
            match="The scope must be 'Library', 'Project, or 'Scenario'"):
        myLibrary.datasheets(scope="test")
        
    with pytest.raises(
            RuntimeError,
            match="The data sheet does not exist: stsim_test"):
        myLibrary.datasheets(name="test")
        
    with pytest.raises(
            ValueError,
            match="filter column Test not in Datasheet stsim_RunControl"):
        myLibrary.datasheets(name="RunControl", filter_column="Test",
                             filter_value=1)
        
    # Test datasheets method outputs
    assert isinstance(myLibrary.datasheets(), pd.DataFrame)
    assert isinstance(myLibrary.datasheets(name="core_Backup"), pd.DataFrame)
    assert isinstance(myLibrary.datasheets(summary=False), list)
    assert len(myLibrary.datasheets().columns) == 3
    assert len(myLibrary.datasheets(optional=True).columns) == 6
    assert myLibrary.datasheets(name="core_Backup", empty=True).empty
    assert not myLibrary.datasheets().equals(
        myLibrary.datasheets(scope="Project"))
    assert not myLibrary.datasheets().equals(
        myLibrary.datasheets(scope="Scenario"))
    
def test_library_delete():
    
    myLibrary = ps.library(name="Test", overwrite=True)
    myLibrary.projects(name="test")
    
    # Test delete method
    with pytest.raises(
            TypeError,
            match="project must be a Project instance, Integer, or String"):
        myLibrary.delete(project=1.5)
    
    with pytest.raises(
            TypeError,
            match="scenario must be a Scenario instance, Integer, or String"):
        myLibrary.delete(scenario=1.5)
        
    with pytest.raises(TypeError, match="force must be a Logical"):
        myLibrary.delete(force="True")
        
    with pytest.raises(ValueError, match="Project ID 2 does not exist"):
        myLibrary.delete(project=2)
        
    with pytest.raises(ValueError, match="project dne does not exist"):
        myLibrary.delete(project="dne")
        
    with pytest.raises(ValueError, match="Scenario ID 50 does not exist"):
        myLibrary.delete(scenario=50)
        
    with pytest.raises(ValueError, match="scenario dne does not exist"):
        myLibrary.delete(scenario="dne")
        
    myLibrary.delete(project="test", force=True)
    assert myLibrary._Library__projects.empty
    assert "test" not in myLibrary.projects().Name.values
    
    myLibrary.scenarios(name="test")
    myLibrary.delete(scenario="test", force=True)
    assert "test" not in myLibrary.scenarios().Name.values
    
def test_library_save_datasheet():
    
    myLibrary = ps.library(name="Test")
    
    # Test save_datasheet method
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument:"):
        myLibrary.save_datasheet(name=1)
    
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument:"):
        myLibrary.save_datasheet(data=1)
        
    with pytest.raises(TypeError, match="name must be a String"):
        myLibrary.save_datasheet(name=1, data=1)
        
    with pytest.raises(TypeError, match="data must be a pandas DataFrame"):
        myLibrary.save_datasheet(name="test", data=1)
        
    with pytest.raises(TypeError, match="scope must be a String"):
        myLibrary.save_datasheet(name="test", data=pd.DataFrame(), scope=1)
        
    with pytest.raises(RuntimeError,
                       match="The data sheet does not exist: stsim_test"):
        myLibrary.save_datasheet(name="test", data=pd.DataFrame())
        
    with pytest.raises(
            RuntimeError,
            match="The header references a column that does not belong"):
        random_df = pd.DataFrame({"col1": [1], "col2": [2]})
        myLibrary.save_datasheet(name="core_Backup", data=random_df)
    
    initial_core_backup = myLibrary.datasheets(name="core_Backup")
    assert initial_core_backup["IncludeOutput"].isna().values[0]
    
    initial_core_backup["IncludeOutput"] = "Yes"
    myLibrary.save_datasheet(name="core_Backup", data=initial_core_backup)
    modified_core_backup = myLibrary.datasheets(name="core_Backup")
    assert (modified_core_backup["IncludeOutput"] == "Yes").item()   
    
def test_library_run():
    
    mySession = ps.Session()
    mySession.add_packages("helloworldSpatial")
    myLibrary = ps.library(name="Test", package="helloworldSpatial",
                           template="example-library", overwrite=True)
    
    # Test run method
    with pytest.raises(
            TypeError,
            match="must be Scenario instance, String, Integer, or List"):
        myLibrary.run(scenarios=pd.DataFrame())

    with pytest.raises(
            TypeError,
            match="project must be Project instance, String, or Integer"):
        myLibrary.run(project=[1])
        
    with pytest.raises(
            TypeError,
            match="jobs must be an Integer"):
        myLibrary.run(jobs="1")
    
    runcontrol = myLibrary.datasheets("RunControl", True, False, False,
                                      "Scenario", None, None, False, 1)
    runcontrol["MaximumIteration"] = 2
    runcontrol["MaximumTimestep"] = 2
    myLibrary.save_datasheet("RunControl", runcontrol, "Scenario", 1)
    
    myLibrary.run()
    assert len(myLibrary.scenarios()) == 2
    assert myLibrary.scenarios().iloc[1]["IsResult"] == "Yes"
    
    myLibrary.run(project=1)
    assert len(myLibrary.scenarios()) == 3
    assert myLibrary.scenarios().iloc[2]["IsResult"] == "Yes"
    
    myLibrary.run(project=1, scenarios=1)
    assert len(myLibrary.scenarios()) == 4     
    assert myLibrary.scenarios().iloc[3]["IsResult"] == "Yes"
    
    myLibrary.projects(name="New Project")
    with pytest.raises(
            ValueError,
            match="Must specify project when > 1 Project in the Library"):
        myLibrary.run()
        
def test_library_addons_functions():
    
    myLibrary = ps.library(name="stsim_test", package="stsim", overwrite=True)
    
    # Test enable_addons method
    with pytest.raises(TypeError,
                       match="name must be a String or List of Strings"):
        myLibrary.enable_addons(name=1)
    
    with pytest.raises(TypeError,
                       match="all elements in name must be Strings"):
        myLibrary.enable_addons(name=[1, True])
        
    myLibrary.enable_addons("stsimsf")
    stsimsf_info = myLibrary.addons[myLibrary.addons["Name"] == "stsimsf"]
    assert stsimsf_info["Enabled"].item() == "Yes"    
    
    # Test disable_addons method
    with pytest.raises(TypeError,
                       match="name must be a String or List of Strings"):
        myLibrary.disable_addons(name=1)
    
    with pytest.raises(TypeError,
                       match="all elements in name must be Strings"):
        myLibrary.disable_addons(name=[1, True])
        
    myLibrary.disable_addons("stsimsf")
    stsimsf_info = myLibrary.addons[myLibrary.addons["Name"] == "stsimsf"]
    assert stsimsf_info["Enabled"].item() == "No" 
    
def test_project_attributes():
    
    myLibrary = ps.library(name="Test", package="helloworldSpatial",
                           overwrite=True)
    myProject = myLibrary.projects(name="Definitions")
    
    # Check attributes
    assert isinstance(myProject.pid, int)
    assert myProject.pid == 1
    assert isinstance(myProject.name, str)
    assert myProject.name == "Definitions"
    assert isinstance(myProject.library, ps.Library)
    
def test_project_scenarios():  
    
    myLibrary = ps.library(name="Test", package="helloworldSpatial",
                           overwrite=True)
    myProject = myLibrary.projects(name="Definitions")
    
    # Test scenarios method
    with pytest.raises(
            TypeError,
            match="name must be a String, Integer, or List of these"):
        myProject.scenarios(name=pd.DataFrame())
        
    with pytest.raises(TypeError, match="sid must be an Integer"):
        myProject.scenarios(sid="1")
        
    with pytest.raises(TypeError, match="optional must be a Logical"):
        myProject.scenarios(optional=1)
        
    with pytest.raises(TypeError, match="summary must be a Logical"):
        myProject.scenarios(summary="1")
      
    assert isinstance(myProject.scenarios(), pd.DataFrame)
    assert isinstance(myProject.scenarios(summary=False), list)
    assert myProject.scenarios().empty
    assert len(myProject.scenarios().columns) == 4
    assert len(myProject.scenarios(optional=True).columns) == 11    
    assert isinstance(myProject.scenarios(name="test"), ps.Scenario)
    assert len(myProject.scenarios()) == 1
    assert isinstance(myProject.scenarios(sid=1), ps.Scenario)    

def test_project_datasheets():

    myLibrary = ps.library(name="Test", package="helloworldSpatial")
    myProject = myLibrary.projects(name="Definitions")
    
    # Test Datasheets
    with pytest.raises(TypeError, match="name must be a String"):
        myProject.datasheets(name=1)
        
    with pytest.raises(TypeError, match="summary must be a Logical or 'CORE'"):
        myProject.datasheets(summary="1")
        
    with pytest.raises(TypeError, match="optional must be a Logical"):
        myProject.datasheets(optional=1)
        
    with pytest.raises(TypeError, match="empty must be a Logical"):
        myProject.datasheets(empty=1)
        
    assert isinstance(myProject.datasheets(), pd.DataFrame)
    assert isinstance(myProject.datasheets(summary=False), list)
    assert myProject.datasheets(
        summary='CORE')["Name"].iloc[0].startswith("core")
    assert len(myProject.datasheets().columns) == 3
    assert len(myProject.datasheets(optional=True).columns) == 6
    assert myProject.datasheets(name="core_Transformer").empty is False
    assert myProject.datasheets(name="core_Transformer", empty=True).empty
    
def test_project_save_datasheet():
    
    myLibrary = ps.library(name="Test", package="helloworldSpatial")
    myProject = myLibrary.projects(name="Definitions")
    
    # Test save_datasheet
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument"):
        myProject.save_datasheet(name=1)
        
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument:"):
        myProject.save_datasheet(data=1)
        
    with pytest.raises(TypeError, match="name must be a String"):
        myProject.save_datasheet(name=1, data=1)
        
    with pytest.raises(TypeError, match="data must be a pandas DataFrame"):
        myProject.save_datasheet(name="test", data=1)
        
    assert myProject.datasheets(name="core_AutoGenTag").empty
    test_datasheet = pd.DataFrame({"Name": ["test"], "AutoGenTagKey": ["key"],
                                  "AutoGenTagValue": [1]})
    myProject.save_datasheet(name="core_AutoGenTag", data=test_datasheet)
    assert myProject.datasheets(name="core_AutoGenTag").empty is False

def test_project_copy_delete():
    
    myLibrary = ps.library(name="Test", package="helloworldSpatial")
    myProject = myLibrary.projects(name="Definitions")

    # Test copy
    with pytest.raises(TypeError, match="name must be a String"):
        myProject.copy(name=1)
        
    myNewProj = myProject.copy()
    assert myNewProj.name == "Definitions - Copy"
    assert myNewProj.datasheets(name="core_AutoGenTag").empty is False
    
    myNewerProj = myProject.copy(name="Definitions 2")
    assert myNewerProj.name == "Definitions 2"
    assert myNewerProj.datasheets(name="core_AutoGenTag").empty is False
    
    # Test delete    
    with pytest.raises(
            TypeError,
            match="scenario must be a Scenario instance, Integer, or String"):
        myNewProj.delete(scenario=[1])  
        
    with pytest.raises(RuntimeError, match="The project does not exist"):
        myNewProj.delete(force=True)
        myNewProj.scenarios()
    

def test_scenarios_attributes():

    myLibrary = ps.library(name="Test", package="helloworldSpatial",
                           overwrite=True)
    myScenario = myLibrary.scenarios("Test Scenario")
    
    # Check attributes
    assert isinstance(myScenario.sid, int) or isinstance(
        myScenario.sid, np.int64)
    assert myScenario.sid == 1
    assert isinstance(myScenario.name, str)
    assert myScenario.name == "Test Scenario"
    assert isinstance(myScenario.project, ps.Project)
    assert isinstance(myScenario.library, ps.Library)
    assert isinstance(myScenario.is_result, str)
    assert myScenario.is_result == "No"
    assert math.isnan(myScenario.parent_id)
    
def test_scenario_datasheets():
    
    myLibrary = ps.library(name="ds_test", package="helloworldSpatial",
                           overwrite=True, template="example-library")
    myScenario = myLibrary.scenarios(sid=1)
    
    # Test datasheets
    with pytest.raises(TypeError, match="name must be a String"):
        myScenario.datasheets(name=1)
        
    with pytest.raises(TypeError, match="summary must be a Logical or 'CORE'"):
        myScenario.datasheets(summary="1")
        
    with pytest.raises(TypeError, match="optional must be a Logical"):
        myScenario.datasheets(optional=1)
        
    with pytest.raises(TypeError, match="empty must be a Logical"):
        myScenario.datasheets(empty=1)
        
    assert isinstance(myScenario.datasheets(), pd.DataFrame)
    assert isinstance(myScenario.datasheets(summary=False), list)
    assert myScenario.datasheets(
        summary='CORE')["Name"].iloc[0].startswith("core")
    assert len(myScenario.datasheets().columns) == 3
    assert len(myScenario.datasheets(optional=True).columns) == 6
    assert isinstance(myScenario.datasheets(
        name="RunControl",
        filter_column="MinimumIteration", 
        filter_value="1"), pd.DataFrame)
    assert len(myScenario.datasheets(
        name="RunControl",
        filter_column="MinimumIteration",
        filter_value="1") == 1)
    assert myScenario.datasheets(
        name="RunControl",
        filter_column="MinimumIteration",
        filter_value="2").empty    
    assert myScenario.datasheets(name="InputDatasheet").empty is False
    assert myScenario.datasheets(name="InputDatasheet", empty=True).empty

def test_scenario_save_datasheet():

    myLibrary = ps.library(name="ds_test", package="helloworldSpatial",
                           overwrite=True, template="example-library")
    myScenario = myLibrary.scenarios(sid=1)    

    # Test save_datasheet
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument"):
        myScenario.save_datasheet(name=1)
        
    with pytest.raises(
            TypeError,
            match="missing 1 required positional argument:"):
        myScenario.save_datasheet(data=1)
        
    with pytest.raises(TypeError, match="name must be a String"):
        myScenario.save_datasheet(name=1, data=1)
        
    with pytest.raises(TypeError, match="data must be a pandas DataFrame"):
        myScenario.save_datasheet(name="test", data=1)
        
    runcontrol = myScenario.datasheets(name="RunControl")
    runcontrol["MaximumIteration"] = 2
    runcontrol["MaximumTimestep"] = 2
    myScenario.save_datasheet("RunControl", runcontrol)

    assert myScenario.datasheets(
        name="RunControl")["MaximumIteration"].item() == 2
    assert myScenario.datasheets(
        name="RunControl")["MaximumTimestep"].item() == 2
    
def test_scenario_run_and_results():
    
    myLibrary = ps.library(name="Test", overwrite=True,
                           package="helloworldSpatial",
                           template="example-library")
    myScenario = myLibrary.scenarios(sid=1)
    runcontrol = myScenario.datasheets(name="RunControl")
    runcontrol["MaximumIteration"] = 2
    runcontrol["MaximumTimestep"] = 2
    myScenario.save_datasheet("RunControl", runcontrol)
    
    # Test run
    with pytest.raises(TypeError, match="jobs must be an Integer"):
        myScenario.run(jobs="1")
        
    myScenario.run(jobs=2)
    assert len(myLibrary.scenarios()) == 2 
    assert myLibrary.scenarios().iloc[1]["IsResult"] == "Yes"
    
    # Test results
    with pytest.raises(TypeError, match="Scenario ID must be an Integer"):
        myScenario.results(sid="5")
        
    with pytest.raises(ValueError, match="not a Results Scenario"):
        myScenario.results(sid=1)
        
    assert isinstance(myScenario.results(), pd.DataFrame)
    assert (myScenario.results()["IsResult"] == "Yes").all()
    res_sid = myLibrary.scenarios().iloc[1]["ScenarioID"].item()
    assert isinstance(myScenario.results(sid=res_sid), ps.Scenario) 
    
    # Test run_log
    myResultsScenario = myScenario.results(sid=res_sid)
    assert isinstance(myResultsScenario.run_log(), pd.DataFrame)
    assert not isinstance(myScenario.run_log(), pd.DataFrame)
    
    # Test datasheet_rasters
    with pytest.raises(TypeError, match="datasheet must be a String"):
        myResultsScenario.datasheet_rasters(datasheet=1, column="test")
        
    with pytest.raises(TypeError, match="column must be a String"):
        myResultsScenario.datasheet_rasters(datasheet="test", column=1)
        
    with pytest.raises(TypeError, match="iteration must be an Integer"):
        myResultsScenario.datasheet_rasters(datasheet="test", column="test",
                                           iteration="test")
        
    with pytest.raises(TypeError, match="timestep must be an Integer"):
        myResultsScenario.datasheet_rasters(datasheet="test", column="test",
                                           timestep="test")
        
    with pytest.raises(ValueError,
                       match="Scenario must be a Results Scenario"):
        myScenario.datasheet_rasters(datasheet="test", column="test")
        
    with pytest.raises(RuntimeError,
                       match="The data sheet does not exist"):
        myResultsScenario.datasheet_rasters(datasheet="test", column="test")
        
    with pytest.raises(ValueError,
                       match="No raster columns found in Datasheet"):
        myResultsScenario.datasheet_rasters(datasheet="OutputDatasheet")
    
    with pytest.raises(
            ValueError,
            match="Column test not found in Datasheet"):
        myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                           column="test")
        
    with pytest.raises(
            ValueError, 
            match="Specified iteration above range of plausible values"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          iteration=3) 
       
    with pytest.raises(ValueError, match="iteration cannot be below 1"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          iteration=0)
       
    with pytest.raises(ValueError,
                       match="Some iteration values outside of range"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          iteration=[1, 2, 3])  
       
    with pytest.raises(
            ValueError, 
            match="Specified timestep above range of plausible values"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          timestep=3) 
       
    with pytest.raises(
            ValueError, 
            match="Specified timestep below range of plausible values"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          timestep=0) 
       
    with pytest.raises(ValueError,
                       match="Some timestep values outside of range"):
       myResultsScenario.datasheet_rasters(datasheet="IntermediateDatasheet",
                                          column="OutputRasterFile",
                                          timestep=[1, 2, 3])  
       
    with pytest.raises(
            ValueError, 
            match = "Must specify a filter_value to filter the filter_column"):
        myResultsScenario.datasheet_rasters(
            datasheet="IntermediateDatasheet",
            column = None,
            filter_column="IntermediateDatasheetID") 
        
    with pytest.raises(
            ValueError, 
            match = "filter column test not in Datasheet"
            ):
        myResultsScenario.datasheet_rasters(
            datasheet="IntermediateDatasheet",
            column = None,
            filter_column="test",
            filter_value="test") 
       
    with pytest.raises(
            ValueError, 
            match="filter_value test does not exist in filter_column"):
        myResultsScenario.datasheet_rasters(
            datasheet="IntermediateDatasheet",
            column = None,
            filter_column="IntermediateDatasheetID",
            filter_value="test") 
      
    raster1 = myResultsScenario.datasheet_rasters(
        datasheet="IntermediateDatasheet", column="OutputRasterFile",
        iteration=1, timestep=1)
    assert isinstance(raster1, ps.Raster)
    
    raster2 = myResultsScenario.datasheet_rasters(
        datasheet="IntermediateDatasheet", column="OutputRasterFile")
    assert len(raster2) == 4
    assert all([isinstance(x, ps.Raster) for x in raster2])
    
    raster3 = myResultsScenario.datasheet_rasters(
        datasheet = "IntermediateDatasheet", 
        column = None,
        filter_column="IntermediateDatasheetID",
        filter_value=2)
    assert isinstance(raster3, ps.Raster)
    
    # Test raster class attributes
    assert os.path.isfile(raster1.source)
    assert isinstance(raster1.name, str)
    assert raster1.name.endswith(".it1.ts1")
    assert isinstance(raster1.dimensions, dict)
    assert all([
        x in raster1.dimensions.keys() for x in [
            "height", "width", "cells"]])
    assert isinstance(raster1.extent, dict)
    assert all([
        x in raster1.extent.keys() for x in [
            "xmin", "xmax", "ymin", "ymax"]])    
    assert isinstance(raster1.crs, rasterio.crs.CRS)
    assert isinstance(raster1.values(), np.ndarray)
    assert isinstance(raster1.values(band=1), np.ndarray)
    
def test_scenario_copy_dep_delete():
    
    myLibrary = ps.library(name="Test", package="helloworldSpatial",
                           overwrite=True, template="example-library")
    myScenario = myLibrary.scenarios(name="My Scenario")
    runcontrol = myScenario.datasheets(name="RunControl")
    runcontrol["MaximumIteration"] = 2
    runcontrol["MaximumTimestep"] = 2
    myScenario.save_datasheet("RunControl", runcontrol)
    
    # Test copy
    with pytest.raises(TypeError, match="name must be a String"):
        myScenario.copy(name=1)
        
    myNewScn = myScenario.copy()
    assert myNewScn.name == "My Scenario - Copy"
    assert myNewScn.datasheets(name="RunControl").empty is False
    assert myNewScn.datasheets(
        name="RunControl")["MaximumIteration"].item() == 2
    
    myNewerScn = myScenario.copy(name="My Scenario 2")
    assert myNewerScn.name == "My Scenario 2"
    assert myNewerScn.datasheets(name="RunControl").empty is False
    assert myNewerScn.datasheets(
        name="RunControl")["MaximumIteration"].item() == 2
    
    # Test dependencies
    with pytest.raises(
            TypeError,
            match="dependency must be a Scenario, String, Integer, or List"):
        myNewScn.dependencies(dependency=1.5)
        
    with pytest.raises(
            TypeError,
            match="remove must be a Logical"):
        myNewScn.dependencies(dependency=myNewerScn.sid, remove="True")
    
    assert myNewScn.dependencies().empty is True
    myNewScn.dependencies(dependency=myNewerScn)
    assert myNewScn.dependencies().empty is False
    assert myNewScn.dependencies().Name.item() == "My Scenario 2"
    
    myNewScn.dependencies(dependency=myNewerScn.sid, remove=True, force=True)
    assert myNewScn.dependencies().empty is True
    
    myNewScn.dependencies(dependency=myNewerScn.name)
    assert myNewScn.dependencies().empty is False
    
    sameNameScn = myScenario.copy(name="My Scenario 2")
    with pytest.raises(
            ValueError,
            match="dependency name not unique, use ID or Scenario"):
        myNewScn.dependencies(dependency=sameNameScn.name)
        
    # Test ignore_dependencies
    with pytest.raises(TypeError, match="value must be a String"):
        myNewScn.ignore_dependencies(value=1)
    
    assert math.isnan(myNewScn.ignore_dependencies())
    
    myNewScn.ignore_dependencies(value="RunControl")
    myNewScn.ignore_dependencies()
    
    assert myNewScn.ignore_dependencies() == "'RunControl'"
    
    myNewScn.ignore_dependencies(value="InputDatasheet,OutputDatasheet")
    assert myNewScn.ignore_dependencies() == "'InputDatasheet,OutputDatasheet'"
    
    # Test merge_dependencies
    with pytest.raises(TypeError, match="value must be a Logical"):
        myNewScn.merge_dependencies(value=1)
    
    assert myNewScn.merge_dependencies() == "No"
    
    myNewScn.merge_dependencies(value=True)
    assert myNewScn.merge_dependencies() == "Yes"
    
    # Test delete            
    with pytest.raises(RuntimeError, match="The scenario does not exist"):
        myNewScn.delete(force=True)
        myNewScn.run()
    


import os
import io
import subprocess
import pandas as pd
import pysyncrosim as ps

class Session(object):
    """
    A class to represent a SyncroSim Session.
    
    """
    __pkgs = None
    
    def __init__(self, location=None, silent=True, print_cmd=False):
        """
        Initializes a pysyncrosim Session instance.

        Parameters
        ----------
        location : Str, optional
            Filepath to SyncroSim executable. If None, then uses default
            location on windows. The default is None.
        silent : Logical, optional
            If True, will not print warnings from the console. The default is 
            True.
        print_cmd : Logical, optional
            If True, arguments from the console command will be printed. The 
            default is False.

        Raises
        ------
        ValueError
            Raises error if the location given does not exist.

        Returns
        -------
        None.

        """
        self.__location = self.__init_location(location)
        self.console_exe = self.__init_console(console=True)
        self.__silent = silent
        self.__print_cmd = print_cmd
        self.__pkgs = self.packages()
     
    @property
    def location(self):
        """
        Retrieves the location for this Session.

        Returns
        -------
        String
            Filepath to SyncroSim Session.
            
        """
        return self.__location
    
    @property
    def silent(self):
        """
        Gets or sets the silent status for this Session.

        Returns
        -------
        Logical
            Silent status.
            
        """
        return self.__silent
    
    @silent.setter
    def silent(self, value):
        if value is None:
            raise AttributeError("The silent status cannot be None.")
        elif not isinstance(value, bool):
            raise AttributeError("silent must be a Logical")
        self.__silent = value
        
    @property
    def print_cmd(self):
        """
        Gets or sets the print_cmd status of the SyncroSim Session.

        Returns
        -------
        Logical
            print_cmd status.

        """
        
        return self.__print_cmd
    
    @print_cmd.setter
    def print_cmd(self, value):
        if value is None:
            raise AttributeError("The print_cmd status cannot be None.")
        elif not isinstance(value, bool):
            raise AttributeError("print_cmd must be a Logical")
        self.__print_cmd = value
        
        
    def version(self):
        """
        Retrieves SyncroSim version.

        Returns
        -------
        String
            Version number.

        """
        args = ["--version"]
        v = self.__call_console(args, decode=True)
        
        return v.rstrip()
    
    def packages(self, installed=True):
        """
        Retrieves DataFrame of installed packages.
        
        Parameters
        ----------
        installed : Logical or String, optional
            If False, then shows all available packages. If "BASE", only shows
            installed base packages. If True, then shows installed 
            addons in addition to base packages. The default is True.

        Returns
        -------
        pkgs : pandas.DataFrame
            DataFrame listing the Name, Display Name, and Version of all
            installed base packages.

        """
        if not isinstance(installed, bool) and installed != "BASE":
            raise TypeError("installed must be Logical or 'BASE'")
        
        if installed is True or installed == "BASE":
            args = ["--list", "--basepkgs"]
            self.__pkgs = self.__call_console(args, decode=True, csv=True)
            self.__pkgs = pd.read_csv(io.StringIO(self.__pkgs))
            
        if installed is True:    
            args = ["--list", "--addons"]
            addons = self.__call_console(args, decode=True, csv=True)
            addons = pd.read_csv(io.StringIO(addons))
            self.__pkgs = self.__pkgs.append(addons).reset_index()
            
        if installed is False:
            self.console_exe = self.__init_console(pkgman=True)
            try:
                args = ["--available"]
                pkgs = self.__call_console(args, decode=True, csv=True)
                return pd.read_csv(io.StringIO(pkgs))
            finally:
                self.console_exe = self.__init_console(console=True)

        return self.__pkgs        
    
    def add_packages(self, packages):
        """
        Installs a package.
        
        Parameters
        ----------
        packages : List or String
            Name or list of names of packages to install. Can also be a 
            filepath (or list of filepaths) to a local package.

        Returns
        -------
        None.

        """
        # Unit tests for inputs
        if not isinstance(packages, str):
            if not isinstance(packages, list):
                raise TypeError("packages must be a String or List")
            elif all(isinstance(pkg, str) for pkg in packages) is False:
                raise TypeError("all packages must be Strings")            
        
        # Add some checks to see whether package is installed
        installed = self.packages()
        installed = installed["Name"].values
        
        # Set executable to package manager
        self.console_exe = self.__init_console(pkgman=True)
        
        exception = True
        try:
        
            if not isinstance(packages, list):
                packages = [packages]
                
            for pkg in packages:
                if pkg in installed:
                    print(f'{pkg} already installed')
                    continue
                
                if os.path.exists(pkg):
                    args = ["--finstall=%s" % pkg]
                else:
                    args = ["--install=%s" % pkg]
                self.__call_console(args)
                
            # Reset packages
            self.console_exe = self.__init_console(console=True)
            self.__pkgs = self.packages()
                
        except RuntimeError as e:
            print(e)
            
        else:
            exception = False
        
        finally:
            if exception is False:
                print(f"{packages} installed successfully")
            # Set executable back to console
            self.console_exe = self.__init_console(console=True)
    
    def remove_packages(self, packages):
        """
        Uninstalls a package.

        Parameters
        ----------
        packages : List or String
            Name or list of names of packages to uninstall.

        Returns
        -------
        None.

        """
        # Unit tests for inputs
        if not isinstance(packages, str):
            if not isinstance(packages, list):
                raise TypeError("packages must be a String or List")
            elif all(isinstance(pkg, str) for pkg in packages) is False:
                raise TypeError("all packages must be Strings")   
        
        # Add some checks to see whether package is installed
        installed = self.packages()
        installed = installed["Name"].values
        
        # Set executable to package manager
        self.console_exe = self.__init_console(pkgman=True)
        
        exception = True
        try:
        
            if not isinstance(packages, list):
                packages = [packages]
                
            for pkg in packages:
                if pkg not in installed:
                    print(f'{pkg} not installed')
                    continue
                args = ["--uninstall=%s" % pkg]
                self.__call_console(args)
                
            # Reset packages
            self.console_exe = self.__init_console(console=True)
            self.__pkgs = self.packages()
            
        except RuntimeError as e:
            print(e)
            
        else:
            exception = False
        
        finally:
            if exception is False:
                print(f"{packages} removed successfully")
            # Set executable back to console
            self.console_exe = self.__init_console(console=True)
    
    def update_packages(self, packages=None):
        """
        Updates a package to the newest version.
        
        Parameters
        ----------
        packages : List or String
            Name or list of names of packages to update. If None, then
            updates all packages. Default is None.

        Returns
        -------
        None.

        """
        # Unit tests for inputs
        if packages is not None:
            if not isinstance(packages, str):
                if not isinstance(packages, list):
                    raise TypeError("packages must be a String or List")
                elif all(isinstance(pkg, str) for pkg in packages) is False:
                    raise TypeError("all packages must be Strings")   
        
        # Add some checks to see whether package is installed
        pkg_df = self.packages()
        installed = pkg_df["Name"].values
        
        self.console_exe = self.__init_console(pkgman=True)
            
        try:
            
            if packages is None:
                args = ["--updateall"]
                
            elif not isinstance(packages, list):
                packages = [packages]
                
            for pkg in packages:
                if pkg not in installed:
                    print(f'{pkg} not installed')
                    continue
                # Compare versions
                v1 = pkg_df[pkg_df["Name"] == pkg].Version.item()
                args = ["--updatepkg=%s" % pkg]
                self.__call_console(args)
                self.console_exe = self.__init_console(console=True)
                # Also resets packages below
                pkg_df2 = self.packages()
                v2 = pkg_df2[pkg_df2["Name"] == pkg].Version.item()
                if v1 == v2:
                    print(f"{pkg} already up to date")
                if v1 < v2:
                    print(f"{pkg} updated from v{v1} to v{v2}")
                self.console_exe = self.__init_console(pkgman=True)
                
        finally:
        
            # Set executable back to console
            self.console_exe = self.__init_console(console=True)
         
    def __init_location(self, location):
        # Initializes the location of the SyncroSim executable
        self.__location = location # dja  not working as expected,  The self.__location value set here is being over written when the function returns
        e = ps.environment._environment()
        if self.__location is None:
            if e.program_directory.item() is None:
                return "C:/Program Files/SyncroSim"
            else:
                return e.program_directory.item()
        elif not os.path.isdir(self.__location):
            raise ValueError("The location is not valid")
        else :
            return location # dja just added a statement to return the passed in location path assuming all the other tests were passed.
                        
    def __init_console(self, console=None, pkgman=None):
        
        if console is True:
            return os.path.join(self.__location,
                                "SyncroSim.Console.exe")
            
        elif pkgman is True:
            return os.path.join(self.__location,
                                "SyncroSim.PackageManager.exe")
            
        else:
            raise ValueError("No executable assigned")
    
    def __call_console(self, args, csv=False, decode=False):
        final_args = []
        
        final_args.append(self.console_exe)
        final_args += args
        
        if csv:
            final_args += ["--csv"]
            
        if self.__print_cmd:
            print(final_args)
        
        result = subprocess.run(
            final_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        
        if result.returncode !=0:
            if self.silent is False:
                print(final_args, flush=True)
            raise RuntimeError(result.stderr.decode('utf-8'))
            
        if decode is True:
            return result.stdout.decode('utf-8')
        else:    
            return result







# Copyright (c) 2021 Apex Resource Management Solution Ltd. (ApexRMS). All rights reserved.
# MIT License

from setuptools import setup, find_packages
from pysyncrosim._version import __version__

setup(name="pysyncrosim",
      version=__version__,
      description="Python interface to SyncroSim",
      author="ApexRMS",
      author_email="info@apexrms.com",
      url="https://www.apexrms.com",
      packages=find_packages(exclude="tests"))

install_requires=['numpy', 'pandas', 'rasterio']


import pandas as pd
import numpy as np
#import sys
import os
import io
import tempfile
import shutil
import pysyncrosim as ps
from pysyncrosim import helper
from pysyncrosim.environment import _environment

pd.set_option("max_columns", 5)

class Library(object):
    """
    A class to represent a SyncroSim Library. 
        
    """
    __projects = None
    __scenarios = None
    __datasheets = None
    
    def __init__(self, location, session): # dja make the name more intuitive as well as matching the docstring name and internal consistentcy
        """
        Initializes a pysyncrosim Library instance.

        Parameters
        ----------
        location : String
            Filepath to Library location on disk.
        session : Session
            pysyncrosim Session instance.

        Returns
        -------
        None.

        """
        self.__name = os.path.basename(location) # dja Changed to use new parameter name.  Changed to using the os.path function basename as this makes script more understandable and pythonic 
        self.__session = session
        self.__location = location  # dja needed to match the name change in the class definition
        self.__addons = self.__init_addons()
        self.__info = None
        self.__package = None
        self.__owner = None
        self.__date_modified = None
        self.__readonly = None
        # All above attributes get created by below function
        self.__init_info()
        self.__description = self.__init_description()

    @property
    def session(self):
        """
        Retrieves the Session associated with this Library.

        Returns
        -------
        Session
            The Session for this Library.

        """
        return self.__session
    
    @property
    def name(self):
        """
        Retrieves or sets the name of this Library.

        Returns
        -------
        String
            Library name.

        """
        return self.__name
    
    @name.setter
    def name(self, value):
        args = ["--setprop", "--lib=%s" % self.location, "--name=%s" % value]
        self.session._Session__call_console(args)
        self.__name = value
        
    @property
    def location(self):
        """
        Retrieves the file path to this Library.

        Returns
        -------
        String
            Library file path.

        """
        return self.__location
    
    @property
    def package(self):
        """
        Retrieves the package this Library is using.

        Returns
        -------
        String
            Package name.

        """
        return self.__package
    
    @property
    def addons(self):
        """
        Retrieves the addon(s) this Library is using.

        Returns
        -------
        String
            Addon name(s).

        """
        return self.__addons
    
    @property
    def info(self):
        """
        Retrieves information about this Library.

        Returns
        -------
        pd.DataFrame
            General library information.

        """
        return self.__info
    
    @property
    def owner(self):
        """
        Gets or sets the owner of this Library.

        Returns
        -------
        String
            Owner of this Library.

        """
        return self.__owner
    
    @owner.setter
    def owner(self, value):
        if not isinstance(value, str):
            raise AttributeError("owner must be a String")
        self.__owner = value
        args = ["--setprop", "--lib=%s" % self.location, "--owner=%s" % value]
        self.session._Session__call_console(args)
        self.__owner = value
        
    @property
    def readonly(self):
        """
        Gets or sets the read-only status of this Library.

        Returns
        -------
        String
            "yes" if this Project is read-only, "no" otherwise.

        """
        return self.__readonly
    
    @readonly.setter
    def readonly(self, value):
        if not isinstance(value, bool):
            raise AttributeError("readonly must be a Logical")
        elif value is True or value == "yes":
            self.__readonly = "yes"
        elif value is False or value == "no":
            self.__readonly = "no"
        else:
            raise TypeError("value must be a Logical")
        args = ["--setprop", "--lib=%s" % self.location,
                "--readonly=%s" % self.__readonly]
        self.session._Session__call_console(args)
        self.__readonly = value
        
    @property
    def description(self):
        """
        Gets or sets the description for this Library.

        Returns
        -------
        String
            Library description.

        """
        return self.__description
    
    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise AttributeError("description must be a String")
        self.__description = value
        args = ["--setprop", "--lib=%s" % self.location,
                "--description=%s" % value]
        self.session._Session__call_console(args)
        self.__description = self.__init_description()
        
    @property
    def date_modified(self):
        """
        Retrieves the last date this Library was modified.

        Returns
        -------
        String
            Last date modified.

        """
        return self.__date_modified
    
    def projects(self, name=None, pid=None, summary=True, overwrite=False):
        """
        Creates or opens one or more SyncroSim Projects in the Library.

        Parameters
        ----------
        name : String, optional
            Name of Project. The default is None.
        pid : Int, optional
            Project ID. The default is None.
        summary : Logical, optional
            If True, returns a DataFrame of Project information. If False, 
            returns a list of Project objects. Default is True.
        overwrite : Logical, optional
            If True, then overwrites an existing Project. Default is False.

        Returns
        -------
        Project, List of Projects, or pandas.DataFrame
            DataFrame containing Project information, including ID, Name, 
            Owner, Last Modified, and Read Only, or List of Projects if 
            `summary=False`.

        """
        
        # Add unit tests for argument types later
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        if pid is not None and not isinstance(
                pid, int) and not isinstance(pid, np.int64):
            raise TypeError("pid must be an Integer")
        if not isinstance(summary, bool):
            raise TypeError("summary must be a Logical")
        if not isinstance(overwrite, bool):
            raise TypeError("overwrite must be a Logical")
        
        if overwrite is True:
            self.delete(project=name)
        
        self.__init_projects()
        
        # If pid provided, check that pid exists
        if pid is not None:
            if pid not in self.__projects["ProjectID"].values:
                raise ValueError(f"Project ID {pid} does not exist")
            elif name is not None and name != self.__projects[
                    self.__projects["ProjectID"] == pid]["Name"].values[0]:
                raise ValueError(
                    f"Project ID {pid} does not match Project name {name}")
          
        p = self.__get_project(name=name, pid=pid)
        
        if p is None:
            
            if summary is True:
                
                # Return DataFrame of Project information
                if name is not None:
                    return self.__projects[self.__projects.Name == name]
                if pid is not None:
                    return self.__projects[self.__projects.ProjectID == pid]
                else:
                    return self.__projects
                
            if summary is False:
        
                proj_list = []
                
                for i in range(0, len(self.__projects)):
                    
                    proj = ps.Project(self.__projects["ProjectID"].loc[i],
                                      self.__projects["Name"].loc[i],
                                      self)
                    
                    proj_list.append(proj)
    
                return proj_list   
            
        elif p.empty:
             
            # If specified Project does not exist, then create it    
            args = ["--create", "--project", "--lib=%s" % self.__location,
                    "--name=%s" % name]
            self.session._Session__call_console(args)
            
            # Reset Projects
            self.__projects = None
            self.__init_projects()
            p = self.__get_project(name, pid)
            
            # Convert np.int64 to native int
            pid = p["ProjectID"].values[0].tolist()
            
            return ps.Project(pid, p["Name"].values[0], self)
                
        else:
            
            # Convert np.int64 to native int
            pid = p["ProjectID"].values[0].tolist()
             
            return ps.Project(pid, p["Name"].values[0], self)
                
        
    def scenarios(self, name=None, project=None, sid=None, pid=None,
                  overwrite=False, optional=False, summary=None,
                  results=False):
        """
        Retrieves a Scenario or DataFrame of Scenarios in this Library.

        Parameters
        ----------
        name : String, Int, or List of these, optional
            Scenario name. If an integer is given, the value will be parsed as
            as Scenario ID. The default is None.
        project : Project, String, or Int, optional
            Project the Scenario belongs to. If no Project is specified when
            creating a new Scenario, the "Definitions" default Project is used.
            The default is None.
        sid : Int or List of Ints, optional
            Scenario ID. If both name and sid are specified, the sid is used. 
            The default is None.
        pid : Int, optional
            Project ID. The default is None.
        overwrite : Logical, optional
            Overwrites an existing Scenario. The default is False.
        optional : Logical, optional
            Return optional information. The default is False.
        summary : Logical, optional
            When name and sid is None, if True, returns a DataFrame of 
            information on existing Scenarios. Otherwise returns a list of 
            Scenario class instances.
        results : Logical, optional
            Return only a list of Results Scenarios. The default is False.

        Returns
        -------
        Scenario, List of Scenarios, or pandas.DataFrame
            If a name or sid is specified, returns a Scenario class instance.
            If no name or sid is specified, and summary is set to False,
            returns a list of Scenario class instances. If no name or sid is
            specified and summary is set to True, returns a pandas.DataFrame
            of Scenario information for this Library.

        """
        # Check types
        if name is not None and not isinstance(name, str)\
            and not isinstance(name, int)\
                and not isinstance(name, list):
            raise TypeError("name must be a String, Integer, or List of these")
            
        if isinstance(name, list) and isinstance(sid, list):
            name = None
            
        if isinstance(name, list):
           if not all(isinstance(item, int) for item in name)\
               and not all(isinstance(item, str) for item in name):
               raise TypeError("All values in name list must be either" + 
                               " Strings or Integers")
        elif name is not None:
            name = [name]
               
        if project is not None and project.__class__.__name__ != "Project":
            if not isinstance(project, str) and not isinstance(project, int):
                raise TypeError(
                    "project must be Project instance, String, or Integer")
                
        if sid is not None and not isinstance(sid, int)\
            and not isinstance(sid, np.int64)\
                and not isinstance(sid, list):
            raise TypeError("sid must be an Integer or List of Integers")
            
        if isinstance(sid, list):
            if not all(isinstance(item, int) for item in sid):
                raise TypeError("All values in sid list must be Integers")
        elif sid is not None:
            sid = [sid]
                
        if pid is not None and not isinstance(
                pid, int) and not isinstance(pid, np.int64):
            raise TypeError("pid must be an Integer")
            
        if not isinstance(overwrite, bool):
            raise TypeError("overwrite must be a Logical")
            
        if not isinstance(optional, bool):
            raise TypeError("optional must be a Logical")
            
        if not isinstance(summary, bool) and summary is not None:
            raise TypeError("summary must be a Logical or None")
            
        if not isinstance(results, bool):
            raise TypeError("results must be a Logical")
          
        # Set the summary argument
        if summary is None:
            if name is not None or sid is not None:
                summary = False
            else:
                summary = True
        
        if (sid is not None and len(sid) == 1)\
            and (name is not None and len(name) == 1):
                print("Both Scenario ID and name specified, using Scenario ID")
                output = self.__extract_scenario(None, project, sid[0], pid,
                                                 overwrite, optional, summary,
                                                 results)
        elif sid is not None:
            output = [self.__extract_scenario(
                name, project, s, pid, overwrite, optional, summary, results
                ) for s in sid]
        elif name is not None:
            output = [self.__extract_scenario(
                n, project, sid, pid, overwrite, optional, summary, results
                ) for n in name]
        else:
            output = self.__extract_scenario(None, project, None, pid,
                                             overwrite, optional, summary,
                                             results)
        if isinstance(output, list) and len(output) == 1:
            return output[0]
        else:
            return output
        
        
    def datasheets(self, name=None, summary=True, optional=False, empty=False,
                   scope="Library", filter_column=None, filter_value=None,
                   include_key=False, *ids):
        """
        Retrieves a DataFrame of Library Datasheets.

        Parameters
        ----------
        name : String, optional
            Datasheet name. The default is None.
        summary : Logical or String, optional
            Whether to list package Datasheets or core Datasheets. The default
            is True.
        optional : Logical, optional
            Return optional columns. The default is False.
        empty : Logical, optional
            Return an empty Datasheet. The default is False.
        scope : String, optional
            Datasheet scope. Options include "Library", "Project", or 
            "Scenario". The default is "Library".
        filter_column : String
            The column to filter the output Datasheet by. The default is None.
        filter_value : String, Int, or Logical
            The value to filter the filter_column by. The default is None.
        include_key : Logical, optional
            Whether to include the primary key of the Datasheet, corresponding
            to the SQL database. Default is False.

        Returns
        -------
        pandas.DataFrame
            If `optional=False`, then returns a DataFrame of Datasheet 
            information including Package, Name, and Display Name.
            If `optional=True`, also returns Scope, Is Single, and Is Output.

        """
        # Type checks
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a String")
        if not isinstance(summary, bool) and summary != "CORE":
            raise TypeError("summary must be a Logical or 'CORE'")
        if not isinstance(optional, bool):
            raise TypeError("optional must be a Logical")
        if not isinstance(empty, bool):
            raise TypeError("empty must be a Logical")
        if not isinstance(filter_column, str) and filter_column is not None:
            raise TypeError("filter_column must be a String")
        if not isinstance(include_key, bool):
            raise TypeError("include_key must be a Logical")
        
        self.__datasheets = None
        
        # TODO: Check if datasheet exists in desired scope
        
        # Initialize base arguments
        args = ["--export", "--lib=%s" % self.__location]
        
        if scope == "Project" and len(ids) > 0:
            args += ["--pid=%d" % ids]
        
        if scope == "Scenario" and len(ids) > 0:               
            args += ["--sid=%d" % ids]
            
        if empty:
            args += ["--schemaonly"]
            
        if include_key:
            args += ["--includepk"]
         
        # Can only use filter_column arg if name of Datasheet is specified
        if (filter_column is not None) and (name is not None):
            
            # Check that filter_value is not None
            if filter_value is None:
                raise ValueError("Must specify a filter_value to filter the "+
                                 "filter_column by")
            
            # Check if filter_column exists in Datasheet
            name = self.__check_datasheet_name(name)            
            check_args = ["--list", "--columns",
                          "--lib=%s" % self.location, 
                          "--sheet=%s" % name]
            ds_cols = self.__console_to_csv(check_args)
            
            if filter_column not in ds_cols.Name.values:
                raise ValueError(
                    f"filter column {filter_column} not in Datasheet {name}")
               
            tempfile_path = None
            try:
                
                filter_value = int(filter_value)

                
            except ValueError:
                    
                # Initialize Datasheets summary
                # TODO: find out why Library and project scoped datasheets not showing up
                self.__init_datasheets(scope=scope, summary=True)
                
                ds_row = self.__datasheets[self.__datasheets.Name == name]
                
                if ds_row["Is Output"].values[0] == "Yes":
                    input_sheet_name = ds_cols[
                        ds_cols.Name == filter_column].Formula1.values[0]
                    filter_column = "Name" # TODO: Check if this is true for all cases
                    
                else:
                    input_sheet_name = name
                
                tempfile_path = os.path.join(self.location + ".temp",
                                             "temp.csv")
                check_args = ["--export", "--lib=%s" % self.location,
                              "--sheet=%s" % input_sheet_name, 
                              "--file=%s" % tempfile_path, "--includepk",
                              "--valsheets", "--extfilepaths", "--force"]
                
                # Check the scope of the input_sheet_name
                scope_list = ["Project", "Scenario"]
                
                if (self.__datasheets.Name == input_sheet_name).any():
                    
                    input_scope = scope
                    
                    if input_scope == "Project" and len(ids) > 0:
                        check_args += ["--pid=%d" % ids]
                    
                    if input_scope == "Scenario" and len(ids) > 0:               
                        check_args += ["--sid=%d" % ids]
                        
                else:
                    scope_list.remove(scope)
                    input_scope = scope_list[0]
                        
                    if input_scope == "Project" and len(ids) > 0:
                        if scope == "Scenario":
                            pid = self.scenarios(sid=ids[0]).project.pid
                        else:
                            pid = ids[0]
                        check_args += ["--pid=%d" % pid]
                    
                    if input_scope == "Scenario" and len(ids) > 0:               
                        check_args += ["--sid=%d" % ids[0]]
                
                self.session._Session__call_console(check_args)
                input_datasheet = pd.read_csv(tempfile_path)
                
                if (input_datasheet[filter_column] != filter_value).all():
                    raise ValueError(f"filter_value {filter_value} does not "+
                                     f"exist in filter_column {filter_column}")
                
                # Get primary key and ID for filter column/value
                primary_key = input_datasheet.columns[0]
                primary_value = input_datasheet[
                    input_datasheet[filter_column] == filter_value
                    ][primary_key].values[0]
                filter_column = primary_key
                filter_value = primary_value
            
            finally:
                if tempfile_path is not None and os.path.exists(tempfile_path):
                    os.remove(tempfile_path)
                self.__datasheets = None
            
            # If all checks pass, then add filter_column to args
            args += ["--filtercol=%s" % filter_column + "=" + str(filter_value)]
        
        if name is None:
            
            # Return DataFrame of Datasheets
            if summary is True or summary == "CORE":
                self.__init_datasheets(scope=scope, summary=summary)
                if optional is False:
                    self.__datasheets = self.__datasheets.iloc[:, 1:4]
                    
                return self.__datasheets
        
            # Return List of DataFrames
            if summary is False:
            
                self.__init_datasheets(scope=scope, summary=True)
                d_summary = self.__datasheets.copy()
                self.__datasheets = None
                                
                ds_list = []
                
                # Add arguments
                args += ["--sheet"]
                
                for ds in d_summary["Name"]:
                    args[-1] = "--sheet=%s" % ds
                    self.__datasheets = None
                    self.__init_datasheets(scope=scope, summary=False,
                                           name=ds, args=args)
                    ds_list.append(self.__datasheets)
                    
                return ds_list
        
        if name is not None:
            
            # If package is not included in name, add it
            name = self.__check_datasheet_name(name)
            
            # Add arguments
            args += ["--sheet=%s" % name] 
            
            if name.startswith("core"):
                args += ["--includesys"]
                
            self.__init_datasheets(scope=scope, summary=False,
                                   name=name, args=args)
            
            return self.__datasheets
        
    def delete(self, project=None, scenario=None, force=False):
        """
        Deletes a SyncroSim class instance.

        Parameters
        ----------
        project : Project, String, or Int, optional
            If called from a Library class instance, specify the Project to 
            delete. The default is None.
        scenario : Scenario, String, or Int, optional
            If called from a Scenario class instance, specify the Scenario to
            delete. The default is None.
        force : Logical, optional
            If set to True, does not ask user before deleting SyncroSim class
            instance. The default is False.

        Returns
        -------
        None.

        """
        # after delete, can still access the instance attributes - should
        # create function to set to None?
        # Also, should have method to delete list of Projects or Scenarios?
        
        # type checks
        if project is not None and not isinstance(project, ps.Project):
            if not isinstance(project, int) and not isinstance(
                    project, str) and not isinstance(project, np.int64):
                raise TypeError(
                    "project must be a Project instance, Integer, or String")
        if scenario is not None and not isinstance(scenario, ps.Scenario):
            if not isinstance(scenario, int) and not isinstance(
                    scenario, str) and not isinstance(scenario, np.int64):
                raise TypeError(
                    "scenario must be a Scenario instance, Integer, or String")
        if not isinstance(force, bool):
            raise TypeError("force must be a Logical")
        
        if project is None and scenario is None:
            
            helper._delete_library(name = self.location, session=self.session,
                               force=force)
        
        elif project is not None and scenario is None:
            
            # turn project into project class instance if str or int
            if type(project) is int:
                p = self.projects(pid = project)
            if type(project) is str:
                if project in self.__projects["Name"].values:
                    p = self.projects(name = project)
                else:
                    raise ValueError(f'project {project} does not exist')
            if isinstance(project, ps.Project):
                p = project
            
            helper._delete_project(library=self, name=p.name,
                                   pid=p.pid, session=self.session,
                                   force=force)
            
        elif scenario is not None:
            
            # turn scenario into scenario class instance if str or int
            if type(scenario) is int:
                s = self.scenarios(sid = scenario, project = project)
            if type(scenario) is str:
                if scenario in self.__scenarios["Name"].values:
                    s = self.scenarios(name = scenario, project = project)
                else:
                    raise ValueError(f'scenario {scenario} does not exist')
            if isinstance(scenario, ps.Scenario):
                s = scenario
            
            helper._delete_scenario(library=self, project=s.project, 
                                    name=s.name, sid=s.sid,
                                    session=self.session,
                                    force=force)
    
    def save_datasheet(self, name, data, scope="Library", *ids):
        """
        Saves a pandas DataFrane as a SyncroSim Datasheet.

        Parameters
        ----------
        name : String
            Name of the Datasheet.
        data : pandas DataFrame
            DataFrame of Datasheet values.
        scope : String, optional
            Scope of the Datasheet. The default is "Library".
        *ids : Int
            If Project- or Scenario-scoped, requires the Project or Scenario
            IDs.

        Returns
        -------
        None.

        """
        # Type checks
        if not isinstance(name, str):
            raise TypeError("name must be a String")
        if not isinstance(data, pd.DataFrame):
            raise TypeError("data must be a pandas DataFrame")
        if not isinstance(scope, str):
            raise TypeError("scope must be a String")
          
        # Add package name to Datasheet name if not included
        if name.startswith(self.package) is False:
            if name.startswith("core") is False:
                name = self.package + "_" + name
            
        # Check if running in a SyncroSim environment from the user interface
        e = _environment()
        transfer_dir = e.transfer_directory.item()
        
        # If running from user interface, save data to transfer directory
        if transfer_dir is not None:
            fpath = '{}\\SSIM_OVERWRITE-{}.csv'.format(transfer_dir, name)
            data.to_csv(fpath, index=False)
            return
        
        # Otherwise export the data to SyncroSim
        else:
            try:
                temp_folder = tempfile.mkdtemp(prefix="SyncroSim-")
                fpath = '{}\\export.csv'.format(temp_folder)
                data.to_csv(fpath, index=False)

                if not os.path.isfile(fpath):
                    raise RuntimeError(f"file path {fpath} does not exist")

                # Specify import arguments
                args = ["--import", "--lib=%s" % self.location,
                        "--sheet=%s" % name, "--folder=%s" % temp_folder,
                        "--file=%s" % fpath]

                if scope == "Project":
                    args += ["--pid=%d" % ids]

                if scope == "Scenario":
                    args += ["--sid=%d" % ids]

                self.__session._Session__call_console(args)

            finally:
                shutil.rmtree(temp_folder, ignore_errors=True)
            
    def run(self, scenarios=None, project=None, jobs=1,
            copy_external_inputs=False):
        """
        Runs a list of Scenario objects.

        Parameters
        ----------
        scenarios : Scenario, String, Int, or List
            List of Scenrios, SyncroSim Scenario instance, name of Scenario,
            or Scenario ID.
        project : Project, optional
            SyncroSim Project instance, name of Project, or Project ID.
        jobs : Int, optional
            Number of multiprocessors to use. The default is 1.
        copy_external_inputs : Logical, optional
            If False, then a copy of external input files (e.g. GeoTIFF files)
            is not created for each job. Otherwise, a copy of external inputs 
            is created for each job. Applies only when jobs > 1. The default is
            False.

        Returns
        -------
        result_dict : Dictionary
            Dictionary of Results Scenarios.

        """
        # Type checks
        if scenarios is not None and not isinstance(
                scenarios, ps.Scenario) and not isinstance(
                    scenarios, int) and not isinstance(
                    scenarios, np.int64) and not isinstance(
                        scenarios, str) and not isinstance(
                            scenarios, list):
            raise TypeError(
                "scenarios must be Scenario instance, String, Integer, or List")
        if project is not None and not isinstance(
                project, ps.Project) and not isinstance(
                    project, int) and not isinstance(
                    project, np.int64) and not isinstance(project, str):
            raise TypeError(
                "project must be Project instance, String, or Integer")
        if not isinstance(jobs, int) and not isinstance(jobs, np.int64):
            raise TypeError("jobs must be an Integer")
        
        # Collect output in a dictionary
        result_list = []
        
        if scenarios is None:
            
            if project is None:
                
                if len(self.projects()) == 1:
                    project = self.projects(
                        pid=self.__projects.ProjectID.item())
                    
                else:
                    raise ValueError(
                        "Must specify project when > 1 Project in the Library")
            
            # Convert project to Project instance
            if isinstance(project, int):
                project = self.projects(pid=project)
            if isinstance(project, str):
                project = self.projects(name=project)
                
            # Run all Scenarios in a Project
            scenarios = project.scenarios(summary=False)
            
            if not isinstance(scenarios, list):
                scenarios = [scenarios]
            
            result_list = [
                scn.run(
                    jobs=jobs, copy_external_inputs=copy_external_inputs
                    ) for scn in scenarios]
                    
        elif scenarios is not None:
                
            if not isinstance(scenarios, list):
                scenarios = [scenarios]
                
            if isinstance(scenarios[0], int):
                scenario_list = [
                    self.scenarios(sid=scn) for scn in scenarios]
            elif isinstance(scenarios[0], str):
                scenario_list = [
                    self.scenarios(
                        name=scn, project=project) for scn in scenarios]
            else:
                scenario_list = scenarios

            result_list = [scn.run(
                jobs=jobs, copy_external_inputs=copy_external_inputs
                ) for scn in scenario_list]
            
        if len(result_list) == 1:
            return result_list[0]
        else:                
            return result_list
        
    def update(self):
        """
        Updates a SyncroSim Library.

        Returns
        -------
        None.

        """
        try:
            args = ["--update", "--lib=%s" % self.location]
            self.session._Session__call_console(args)
        
        except RuntimeError as e:
            print(e)
        
    def backup(self):
        """
        Creates a backup of a SyncroSim Library.

        Returns
        -------
        None.

        """
        ds = self.datasheets(name = "core_Backup")
        args = ["--lib=%s" % self.location, "--backup"]
        
        if ds.IncludeInput[0] == "Yes":
            args += ["--input"]
            
        if ds.IncludeOutput[0] == "Yes":
            args += ["--output"]
        
        self.session._Session__call_console(args)
        
    def enable_addons(self, name):
        """
        Enable addon package(s) of a SyncroSim Library.

        Parameters
        ----------
        name : String or List
            Name of addon(s).

        Raises
        ------
        ValueError
            addon does not exist in available addons for this Library.

        Returns
        -------
        None.

        """
        
        # Type checks and convert to list
        if not isinstance(name, list):
            if not isinstance(name, str):
                raise TypeError("name must be a String or List of Strings")
            else:
                name = [name]
        elif all(isinstance(x, str) for x in name) is False:
            raise TypeError("all elements in name must be Strings")
            
            
        for a in name:
            args = ["--create", "--addon", "--lib=%s" % self.location,
                    "--name=%s" % a]
            try:
                self.session._Session__call_console(args)
                
            # Convert SyncroSim console error to print output
            except RuntimeError as e:
                print(e)
                return
         
        # Reset addons
        self.__addons = self.__init_addons()
    
    def disable_addons(self, name):
        """
        Disable addon package(s) of a SyncroSim Library.

        Parameters
        ----------
        name : String
            Name of addon package(s).

        Raises
        ------
        ValueError
            addon does not exist in available addons for this Library.

        Returns
        -------
        None.

        """
        
        # Type checks and convert to list
        if not isinstance(name, list):
            if not isinstance(name, str):
                raise TypeError("name must be a String or List of Strings")
            else:
                name = [name]
        elif all(isinstance(x, str) for x in name) is False:
            raise TypeError("all elements in name must be Strings")
            
        for a in name:
            args = ["--delete", "--addon", "--force",
                    "--lib=%s" % self.location, "--name=%s" % a]
            try:
                self.session._Session__call_console(args)
                
            # Convert SyncroSim console error to print output
            except RuntimeError as e:
                print(e)
                return
            
        # Reset addons
        self.__addons = self.__init_addons()
        
    def __init_addons(self):   
        # Retrieves addons information
        args = ["--list", "--addons", "--lib=%s" % self.location]
        return self.__console_to_csv(args)
    
    def __init_info(self):
        # Retrieves Library info
        args = ["--list", "--library", "--lib=%s" % self.location]
        lib_info = self.__console_to_csv(args)
        self.__info = lib_info
        
        # Set index and retrieve all properties
        lib_info = lib_info.set_index("Property")
        self.__package = lib_info.loc["Package Name:"].item()
        self.__owner = lib_info.loc["Owner:"].item()
        self.__readonly = lib_info.loc["Read Only:"].item()
        self.__date_modified = lib_info.loc["Last Modified:"].item()
        
    def __init_description(self):  
        # Retrieves the Library description
        args = ["--list", "--description", "--lib=%s" % self.__location]
        return self.session._Session__call_console(args, decode=True)
                
    def __init_projects(self): 
        # Retrieves a list of Projects
        args = ["--list", "--projects", "--lib=%s" % self.__location]
        self.__projects = self.__console_to_csv(args)
            
    def __init_scenarios(self, pid=None):
        # Retrieves a list of Scenarios
        args = ["--list", "--scenarios", "--lib=%s" % self.__location]
        if pid is not None:
            args += ["--pid=%d" % pid]
        self.__scenarios = self.__console_to_csv(args)
            
    def __init_datasheets(self, scope, summary, name=None, args=None):
        # Retrieves a list of Datasheets
        if self.__datasheets is None:
            if summary is True:
                args = ["--list", "--datasheets", "--lib=%s" % self.__location,
                        "--scope=%s" % scope]
            if summary == "CORE":
                args = ["--list", "--datasheets", "--lib=%s" % self.__location,
                        "--scope=%s" % scope, "--includesys"]
            self.__datasheets = self.__console_to_csv(args)    
            
    def __check_datasheet_name(self, name):
        # Appends package name to Datasheet name
        if name.startswith(self.package) is False:
            if name.startswith("core") is False:
                name = self.package + "_" + name
        return name
            
    def __get_project(self, name=None, pid=None):
        # Retrieves Project info from the name or ID
        if name is None and pid is None:
            return None
        if name is not None:
            return self.__projects[self.__projects["Name"] == name]
        else:
            return self.__projects[self.__projects["ProjectID"] == pid]
            
    def __get_scenario(self, name=None, sid=None):
        # Retrieves Scenario info from the name or ID
        if name is None and sid is None:
            return None
        if name is not None:
            return self.__scenarios[self.__scenarios["Name"] == name]
        else:
            return self.__scenarios[self.__scenarios["ScenarioID"] == sid]
        
    def __extract_scenario(self, name, project, sid, pid, overwrite, optional,
                           summary, results):
        
        # Find out if first argument is a Scenario ID
        if isinstance(name, int):
            
            if sid is not None:
                raise ValueError("Name is specified as a Scenario ID, but " + 
                                 "sid is already given")
                
            sid = name
            name = None
        
        # Find project if not specified
        if summary is False:
            if project is None and pid is None:
                
                self.__init_scenarios()
                if sid is not None and self.__get_scenario(sid=sid).empty is False:
                    pid = self.__get_scenario(sid=sid)["ProjectID"].item()
                    project = self.projects(pid=pid)
                if name is not None and len(self.__get_scenario(name=name)) == 1:
                    pid = self.__get_scenario(name=name)["ProjectID"].item()
                    project = self.projects(pid=pid)
                elif self.__projects is None or self.__projects.empty:
                    project = self.projects(name = "Definitions")
                    pid = project.pid
                elif len(self.__projects) == 1:
                    pid = self.__projects.ProjectID.item()
                    project = self.projects(pid=pid)
                else:
                    raise ValueError("More than one Project in Library." + 
                                     " Please specify a Project.")
                    
            elif isinstance(project, int) or isinstance(project, np.int64):
                pid = project
            elif isinstance(project, str):
                pid = self.__projects[
                    self.__projects.Name == project].ProjectID.item()
            elif isinstance(project, ps.Project):
                pid = project.pid
                
            if overwrite is True:
                self.delete(project=project, scenario=name, force=True)
                    
        # Retrieve Scenario DataFrame
        self.__init_scenarios(pid=pid)
        
        # If sid provided, check that sid exists
        if sid is not None:
            
            if self.__get_scenario(sid=sid).empty:
                raise ValueError(f"Scenario ID {sid} does not exist")
            elif name is not None and name != self.__get_scenario(
                    sid=1).Name.item():
                raise ValueError(
                    f"Scenario ID {sid} does not match Scenario name {name}")
        
        s = self.__get_scenario(name=name, sid=sid)
        
        if (s is None) or (summary is True):
            
            # Retrieve DataFrame of available Scenarios
            if summary:
                
                if optional is False:
                    ds =  self.__scenarios[['ScenarioID',
                                            'ProjectID',
                                            'Name',
                                            'IsResult']]
                else:
                    ds = self.__scenarios
                    
                if results:
                    ds = ds[ds["IsResult"] == "Yes"]
                
                if name is not None:
                    return ds[ds.Name == name]
                
                if sid is not None:
                    return ds[ds["ScenarioID"] == sid]
                
                return ds
              
            # Return list of Scenario objects    
            if summary is False:
                
                s_summary = self.__scenarios
                s_summary = s_summary[s_summary["IsResult"] == "No"]
                
                if not isinstance(project, ps.Project):
                    project = self.projects(pid=pid) 
                                
                scn_list = []
                
                for i in range(0, len(s_summary)):
                    
                    scn = ps.Scenario(s_summary["ScenarioID"].loc[i],
                                      s_summary["Name"].loc[i],
                                      project, self)
                    
                    scn_list.append(scn)
    
                return scn_list                
        
        
        # Create a Scenario
        if s.empty:
            
            if isinstance(project, int):
                pid = project
            elif isinstance(project, str):
                self.__init_projects()
                p = self.__get_project(name=project)
                pid = p["ProjectID"].values[0]
            elif isinstance(project, ps.Project):
                pid = project.pid
             
            # Create Scenario using console    
            args = ["--create", "--scenario", "--pid=%d" % pid,
                    "--lib=%s" % self.__location, "--name=%s" % name]
            self.session._Session__call_console(args)
            
            # Reset Scenarios
            self.__scenarios = None
            self.__init_scenarios()
            s = self.__get_scenario(name=name, sid=sid)
            
            if not isinstance(project, ps.Project):
                project = self.projects(pid=pid)
            
            return ps.Scenario(s["ScenarioID"].values[0],
                               s["Name"].values[0], project, self)
        
        # Open a Scenario
        else:
                            
            # Retrieve the name of a Scenario using only the sid
            pid = s["ProjectID"].values[0].tolist()

            if not isinstance(project, ps.Project):
                project = self.projects(pid=pid)

            sid = s["ScenarioID"].values[0].tolist()
            
            return ps.Scenario(sid, s["Name"].values[0], project, self)

        
    def __console_to_csv(self, args, index_col=None):
        # Turns console output into a pd.DataFrame
        console_output = self.session._Session__call_console(
            args,
            decode=True,
            csv=True)
        
        return pd.read_csv(io.StringIO(console_output), index_col=index_col)
        
        
        
        

